# -*- coding: utf-8 -*-

"""
file: processors.py

Description: formerly sem_modules. The purpose is to improve the code in two
main ways: terminology and organization. "Module" conflicts with python modules
while "processor" is more accepted (plus the "main" method of a "module" was...
process_document). Organization is improved because we do not conflate the
processing object with the "user interface" that made using processors in code
a little bit unwieldy and awkward. This file also replaces pipeline.py

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import time
import tempfile
import re
import sys
from datetime import timedelta

try:
    from xml.etree.cElementTree import ElementTree, tostring as element2string
except ImportError:
    from xml.etree.ElementTree import ElementTree, tostring as element2string

from wapiti.api import Model as WapitiModel

import pymorphy2

import sem.logger
import sem.annotators

from sem.util import (strip_html, read_chunks, check_model_available, longest_common_substring)
from sem.tokenisers import get_tokeniser
from sem.storage import (
    AnnotationSet, Entry, Segmentation, Sentence, Span, Tag, Trie,
    chunk_annotation_from_sentence, compile_map
)
from sem.features import (xml2feat, NUL)
from sem.exporters import get_exporter


DEFAULT_LICENSE = (
    "No license provided.\n"
    "Unless you find contradictory information from an official source,"
    " this is provided for research, teaching and personal use only,"
    " it may not be used for any other purpose."
)


class Processor:
    def __init__(self, pipeline_mode="all", license=None, **kwargs):
        self.pipeline_mode = pipeline_mode
        self._license = license or DEFAULT_LICENSE

    @property
    def license(self):
        return self._license

    @license.setter
    def license(self, license):
        if self._license and self._license != DEFAULT_LICENSE:
            sem.logger.warning("changing non default license.")
        self._license = license

    def check_mode(self, expected_mode):
        return expected_mode == "all" or self.pipeline_mode in ("all", expected_mode)

    def process_document(self, document, **kwargs):
        raise NotImplementedError(
            "process_document not implemented for root type {}".format(self.__class__)
        )


class AnnotateProcessor(Processor):
    def __init__(self, annotator, field, *args, **kwargs):
        super(AnnotateProcessor, self).__init__(**kwargs)

        self._annotator = sem.annotators.get_annotator(annotator).load(field, *args, **kwargs)

    def process_document(self, document, **kwargs):
        start = time.time()
        self._annotator.process_document(document)
        laps = time.time() - start
        sem.logger.info("done in {}".format(timedelta(seconds=laps)))


def token_spans_buffered(tokeniser, content):
    """Return the token spans of content.
    This does the same as tokeniser.word_spans, but this method buffers
    the input to allow a quicker processing of large content.
    """
    rem = ""  # remainder of unsegmented tokens
    shift = 0
    token_spans = []
    for chunk in read_chunks(content):
        chnk = rem + chunk
        spans = tokeniser.word_spans(chnk)
        if not spans:
            rem = chnk
            continue
        elif spans[-1].ub < len(chnk):
            rem = chnk[spans[-1].ub:]
        elif spans[-1].ub == len(chnk):
            rem = chnk[spans[-1].lb:]
            del spans[-1]
        else:
            rem = ""
        token_spans.extend([Span(shift + s.lb, shift + s.ub) for s in spans])
        shift += len(chnk) - len(rem)
        del spans[:]

    if rem:
        spans = tokeniser.word_spans(rem) or [Span(0, len(rem))]
        token_spans.extend([Span(shift + s.lb, shift + s.ub) for s in spans])
    if not content[token_spans[-1].lb: token_spans[-1].ub].strip():
        del token_spans[-1]

    return token_spans


class SegmentationProcessor(Processor):
    def __init__(self, tokeniser, **kwargs):
        super(SegmentationProcessor, self).__init__(**kwargs)

        if isinstance(tokeniser, str):
            sem.logger.info('Getting tokeniser "{0}"'.format(tokeniser))
            self._tokeniser = get_tokeniser(tokeniser)()
        else:
            self._tokeniser = tokeniser

    def process_document(self, document, **kwargs):
        """
        Updates a document with various segmentations and creates
        an sem.corpus (CoNLL-formatted data) using field argument as index.

        Parameters
        ----------
        document : sem.storage.Document
            the input data. It is a document with only a content
        """

        start = time.time()

        current_tokeniser = self._tokeniser

        sem.logger.debug('segmenting "%s" content', document.name)

        content = document.content
        if document.metadata("MIME") == "text/html":
            content = strip_html(content, keep_offsets=True)

        if document.segmentation("tokens") is None:
            token_spans = token_spans_buffered(current_tokeniser, document.content)
            document.add_segmentation(Segmentation("tokens", spans=token_spans))
        else:
            sem.logger.info("{} already has tokens".format(document.name))
            token_spans = document.segmentation("tokens").spans

        if document.segmentation("sentences") is None:
            sentence_spans = current_tokeniser.sentence_spans(content, token_spans)
            document.add_segmentation(
                Segmentation(
                    "sentences", reference=document.segmentation("tokens"), spans=sentence_spans
                )
            )
        else:
            sem.logger.info("{} already has sentences".format(document.name))
            sentence_spans = document.segmentation("sentences").spans

        if document.segmentation("paragraphs") is None:
            paragraph_spans = current_tokeniser.paragraph_spans(
                content, sentence_spans, token_spans
            )
            document.add_segmentation(
                Segmentation(
                    "paragraphs",
                    reference=document.segmentation("sentences"),
                    spans=paragraph_spans,
                )
            )
        else:
            sem.logger.info("{} already has paragraphs".format(document.name))

        if len(document.corpus) == 0:
            document.corpus.from_segmentation(
                document.content,
                document.segmentation("tokens"),
                document.segmentation("sentences"),
            )

        sem.logger.info(
            '"{0}" segmented in {1} sentences, {2} tokens'.format(
                document.name, len(sentence_spans), len(token_spans)
            )
        )

        laps = time.time() - start
        sem.logger.info("in {0}".format(timedelta(seconds=laps)))


class EnrichProcessor(Processor):
    def __init__(
        self,
        path=None,
        bentries=None,
        aentries=None,
        features=None,
        mode="label",
        **kwargs,
    ):
        super(EnrichProcessor, self).__init__(**kwargs)

        self._mode = mode
        self._source = path
        self._bentries = []  # informations that are before newly added information
        self._aentries = []  # informations that are after ...
        self._features = []  # informations that are added
        self._names = set()
        self._temporary = set()  # features that will be deleted at the end of process
        self._x2f = None  # the feature parser, initialised in parse

        if self._source is not None:
            sem.logger.info("loading %s", self._source)
            self._parse(self._source)
        else:
            self._bentries = (
                [entry for entry in bentries if entry.has_mode(self._mode)]
                if bentries
                else self._bentries
            )
            self._aentries = (
                [entry for entry in aentries if entry.has_mode(self._mode)]
                if aentries
                else self._aentries
            )
            self._features = features
            self._names = set([entry.name for entry in self._aentries + self._bentries])

    @property
    def mode(self):
        return self._mode

    @mode.setter
    def mode(self, mode):
        if not isinstance(self._source, str):
            raise RuntimeError(
                "cannot change mode for Enrich module: source for informations is not a file."
            )
        self._mode = mode
        sem.logger.info("loading %s", self._source)
        self._parse(self._source)

    @property
    def bentries(self):
        return self._bentries

    @property
    def aentries(self):
        return self._aentries

    @property
    def features(self):
        return self._features

    def fields(self):
        fields = [entry.name for entry in self._bentries]
        fields += [name for (feature, name) in self.features if name not in self._temporary]
        fields += [entry.name for entry in self._aentries]
        return fields

    def process_document(self, document, **kwargs):
        """
        Updates the CoNLL-formatted corpus inside a document with various
        features.

        Parameters
        ----------
        document : sem.storage.Document
            the input data, contains an object representing CoNLL-formatted
            data. Each token is a dict which works like TSV.
        """

        start = time.time()

        missing_fields = set([entry.name for entry in self.bentries + self.aentries]) - set(
            document.corpus.fields
        )

        if len(missing_fields) > 0:
            raise ValueError(
                "Missing fields in input corpus: {0}".format(",".join(sorted(missing_fields)))
            )

        sem.logger.info('enriching file "%s"', document.name)

        fields = self.fields()
        nth = 0
        for i, p in enumerate(document.corpus):
            p.update(self.features)
            for tmp in self._temporary:
                p.remove(tmp)
            nth += 1
            if 0 == nth % 1000:
                sem.logger.debug("%i sentences enriched", nth)
        sem.logger.debug("%i sentences enriched", nth)
        document.corpus.fields = fields[:]

        laps = time.time() - start
        sem.logger.info("done in %s", timedelta(seconds=laps))

    def _parse(self, filename):
        def check_entry(entry_name):
            if entry_name in self._names:
                raise ValueError('Duplicated column name: "{}"'.format(entry_name))
            else:
                self._names.add(entry_name)

        parsing = ElementTree()
        parsing.parse(filename)

        children = parsing.getroot().getchildren()

        if len(children) != 2:
            raise RuntimeError(
                "Enrichment file requires exactly 2 fields, {0} given.".format(len(children))
            )
        else:
            if children[0].tag != "entries":
                raise RuntimeError(
                    'Expected "entries" as first field, got "{0}".'.format(children[0].tag)
                )
            if children[1].tag != "features":
                raise RuntimeError(
                    'Expected "features" as second field, got "{0}".'.format(children[1].tag)
                )

        entries = list(children[0])
        if len(entries) not in (1, 2):
            raise RuntimeError(
                "Entries takes exactly 1 or 2 fields, {0} given".format(len(entries))
            )
        else:
            entry1 = entries[0].tag.lower()
            entry2 = entries[1].tag.lower() if len(entries) == 2 else None
            if entry1 not in ("before", "after"):
                raise RuntimeError(
                    'For entry position, expected "before" or "after", got "{0}".'.format(entry1)
                )
            if entry2 and entry2 not in ("before", "after"):
                raise RuntimeError(
                    'For entry position, expected "before" or "after", got "{0}".'.format(entry2)
                )
            if entry1 == entry2:
                raise RuntimeError("Both entry positions are the same, they should be different")

        for entry in entries:
            for c in entry.getchildren():
                current_entry = Entry.fromXML(c)
                check_entry(current_entry.name)
                if entry.tag == "before" and current_entry.has_mode(self._mode):
                    self._bentries.append(current_entry)
                elif entry.tag == "after" and current_entry.has_mode(self._mode):
                    self._aentries.append(current_entry)

        features = list(children[1])
        del self._features[:]
        for feature in features:
            feature_name = feature.attrib.get("name")
            if not sem.util.str2bool(feature.attrib.get("display", "yes")):
                self._temporary.add(feature_name)
            self._features.append((xml2feat(feature, path=filename), feature_name))
            if not feature_name:
                try:
                    raise ValueError("Nameless feature found.")
                except ValueError as exc:
                    for line in element2string(feature).rstrip().split("\n"):
                        sem.logger.error(line.strip())
                    sem.logger.exception(exc)
                    raise
            check_entry(feature_name)


class CleanProcessor(Processor):
    def __init__(self, to_keep, **kwargs):
        super(CleanProcessor, self).__init__(**kwargs)

        if isinstance(to_keep, str):
            self._allowed = to_keep.split(",")  # comma-separated named fields
        else:
            self._allowed = to_keep

        if len(self._allowed) == 0:
            raise ValueError("No more data after cleaning !")

    def process_document(self, document, **kwargs):
        """
        Cleans the sem.storage.corpus of a document, removing unwanted fields.

        Parameters
        ----------
        document : sem.storage.Document
            the document containing the corpus to clean.
        ranges : str or list of int or list of str
            if str: fields to remove will be induced
            if list of int: each element in the list is the index of a field
            to remove in corpus.fields
            if list of string: the list of fields to remove
        """

        start = time.time()

        sem.logger.info("cleaning document")

        allowed = set(self._allowed)
        fields = set(field for field in document.corpus.fields)

        if len(allowed - fields) > 0:
            sem.logger.warning(
                "the following fields are not present in document,"
                " this might cause an error sometime later: {}".format(", ".join(allowed - fields))
            )

        for sentence in document.corpus.sentences:
            sentence = Sentence({key: sentence.feature(key) for key in allowed})
        document._corpus.fields = self._allowed[:]

        laps = time.time() - start
        sem.logger.info("done in {0}".format(timedelta(seconds=laps)))


def model_from_string(mdl_str, encoding="utf-8"):
    with tempfile.NamedTemporaryFile() as fl:
        try:
            fl.write(mdl_str.encode(encoding=encoding))
        except AttributeError:
            fl.write(mdl_str)
        fl.seek(0)
        return WapitiModel(encoding=encoding, model=fl.name)


class WapitiLabelProcessor(Processor):
    def __init__(
        self, model, field, annotation_fields=None, model_str=None, model_encoding="utf-8", **kwargs
    ):
        super(WapitiLabelProcessor, self).__init__(**kwargs)

        if model is not None and model_str is not None:
            raise ValueError("both 'model' and 'model_str' were provided, only one may be given.")

        self._wapiti_model = None
        self._mdl_str = model_str
        self._model_encoding = model_encoding
        self._model = model
        if self._model is not None:
            self._model = str(self._model)
        self._field = field
        self._annotation_fields = annotation_fields
        if type(self._annotation_fields) == str:
            self._annotation_fields = self._annotation_fields.split(",")

        self.load_model()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_wapiti_model']
        state['_wapiti_model'] = None
        return state

    def __setstate__(self, newstate):
        self.__dict__.update(newstate)
        self.load_model()

    @property
    def field(self):
        return self._field

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, model):
        self._model = model
        if self._model is not None:
            self._model = str(self._model)

        check_model_available(self._model)
        self._wapiti_model = WapitiModel(encoding="utf-8", model=self._model)
        with open(self._model, "rb") as input_stream:
            self._mdl_str = input_stream.read()

    def load_model(self):
        if self._mdl_str is not None:
            self._wapiti_model = model_from_string(self._mdl_str, self._model_encoding)
        elif self._model is not None:
            try:
                check_model_available(self._model)
                self._wapiti_model = WapitiModel(encoding="utf-8", model=self._model)
            except FileNotFoundError:
                sem.logger.warning(
                    "Model file {} does not exist, you will need to train one.".format(self._model)
                )
        else:
            sem.logger.warning("No available model to load.")

    def process_document(self, document, encoding="utf-8", **kwargs):
        """
        Annotate document with Wapiti.

        Parameters
        ----------
        document : sem.storage.Document
            the input data. It is a document with only a content
        """

        start = time.time()

        if self._field in document.corpus.fields:
            sem.logger.warning(
                "field %s already exists in document, not annotating", self._field
            )

            tags = [[s[self._field] for s in sentence] for sentence in document.corpus]
            document.add_annotation_from_tags(tags, self._field, self._field)
        else:
            sem.logger.info("annotating document with %s field", self._field)

            self._label_document(document, encoding)

        laps = time.time() - start
        sem.logger.info("in %s", timedelta(seconds=laps))

    def _label_document(self, document, encoding="utf-8"):
        fields = self._annotation_fields or document.corpus.fields
        tags = []
        for sequence in document.corpus:
            tagging = self._tag_as_wrapper(sequence, fields)
            tags.append(tagging[:])

        document.add_annotation_from_tags(tags, self._field, self._field)

    def _tag_as_wrapper(self, sequence, fields, encoding=None):
        encoding = encoding or self._model_encoding
        seq_str = "\n".join(
            "\t".join(str(it) for it in item)
            for item in zip(*[sequence.feature(field) for field in fields])
        ).encode(encoding)
        s = self._wapiti_model.label_sequence(seq_str).decode(encoding)
        return s.strip().split("\n")


def normalize(token):
    normalized = re.sub("[\u2019]", "'", token, flags=re.U)
    normalized = re.sub("[àáâãäåæ]", "a", normalized, flags=re.U)
    normalized = re.sub("[ÀÁÂÃÄÅÆ]", "A", normalized, flags=re.U)
    normalized = re.sub("[éèêë]", "e", normalized, flags=re.U)
    normalized = re.sub("[ÉÈÊË]", "E", normalized, flags=re.U)
    normalized = re.sub("[ìíîï]", "i", normalized, flags=re.U)
    normalized = re.sub("[ÌÍÎÏ]", "I", normalized, flags=re.U)
    return normalized


def abbrev_candidate(token):
    return token.isupper() and all([c.isalpha() for c in token])


def tokens_from_bounds(document, start, end):
    word_spans = document.segmentation("tokens")
    toks = []
    for span in word_spans:
        if span.lb >= start:
            toks.append(document.content[span.lb: span.ub])
            if span.ub >= end:
                break
        if span.lb > end:
            break
    return toks


def tokens2regex(tokens, flags=re.U):
    apostrophes = "['\u2019]"
    apos_re = re.compile(apostrophes, re.U)
    pattern = "{}{}{}".format("\\b", "\\b *\\b".join(tokens), "\\b")
    pattern = apos_re.sub(apostrophes, pattern)
    pattern = re.escape(pattern)
    return re.compile(pattern, flags)


def detect_abbreviations(document, field):
    content = document.content
    word_spans = document.segmentation("tokens")
    if document.segmentation("sentences") is not None:
        sentence_spans = document.segmentation("sentences").spans
        sentence_spans_ref = document.segmentation("sentences").get_reference_spans()
    else:
        sentence_spans_ref = [Span(0, len(document.content))]
    tokens = [content[span.lb: span.ub] for span in word_spans]
    annotations = document.annotationset(field).get_reference_annotations()

    counts = {}
    positions = {}
    for i, token in enumerate(tokens):
        if (
            abbrev_candidate(token)
            and len(token) > 1
            and not (
                (i > 1 and abbrev_candidate(tokens[i - 1]))
                or (i < len(tokens) - 1 and abbrev_candidate(tokens[i + 1]))
            )
        ):
            if token not in counts:
                counts[token] = 0
                positions[token] = []
            counts[token] += 1
            positions[token].append(i)
    position2sentence = {}
    for token, indices in positions.items():
        for index in indices:
            for i, span in enumerate(sentence_spans):
                if span.lb <= index and span.ub >= index:
                    position2sentence[index] = sentence_spans_ref[i]

    reg2type = {}
    for key in counts:
        all_solutions = []
        for position in positions[key]:
            span = position2sentence[position]
            word_span = word_spans[position]
            lb = span.lb
            ub = word_span.lb
            solutions = longest_common_substring(
                content[lb:ub], tokens[position], casesensitive=False
            )
            if solutions == []:
                solutions = longest_common_substring(
                    normalize(content[lb:ub]), tokens[position], casesensitive=False
                )
            solutions = [
                solution for solution in solutions if len(solution) == len(tokens[position])
            ]
            if len(solutions) > 0:
                all_solutions.extend(
                    [[(x + lb, y + lb) for (x, y) in solution] for solution in solutions]
                )
        if len(all_solutions) > 0:
            all_solutions.sort(key=lambda x: x[-1][0] - x[0][0])
            best_solution = all_solutions[0]
            lo = best_solution[0][0]
            hi = best_solution[-1][0]
            lo_tokens = [tok for tok in word_spans if tok.lb <= lo and tok.ub > lo]
            hi_tokens = [tok for tok in word_spans if tok.lb <= hi and tok.ub > hi]
            abbrev_annots = []
            for position in positions[key]:
                span = word_spans[position]
                abbrev_annots.extend(
                    [
                        annotation
                        for annotation in annotations
                        if annotation.lb == span.lb and annotation.ub == span.ub
                    ]
                )
            try:
                toks = tokens_from_bounds(document, lo_tokens[0].lb, hi_tokens[0].ub)
                reg = tokens2regex(toks, re.U + re.I)
                for match in reg.finditer(content):
                    annots = [
                        annotation
                        for annotation in annotations
                        if (
                            (annotation.lb <= match.start() and match.start() <= annotation.ub)
                            or (annotation.lb <= match.end() and match.end() <= annotation.ub)
                        )
                    ]
                    if len(annots) > 0:
                        annot = annots[0]
                        new_toks = tokens_from_bounds(
                            document, min(annot.lb, match.start()), max(annot.ub, match.end())
                        )
                        new_reg = tokens2regex(new_toks, re.U + re.I)
                        if new_reg.pattern not in reg2type:
                            reg2type[new_reg.pattern] = []
                        reg2type[new_reg.pattern].append(annots[0].value)
                        if abbrev_annots == []:
                            abbrev_reg = tokens2regex([key], re.U)
                            if abbrev_reg.pattern not in reg2type:
                                reg2type[abbrev_reg.pattern] = []
                            reg2type[abbrev_reg.pattern].append(annots[0].value)
                if len(abbrev_annots) > 0:
                    tag = abbrev_annots[0]
                    new_reg = tokens2regex(toks, re.U + re.I)
                    if new_reg.pattern not in reg2type:
                        reg2type[new_reg.pattern] = []
                    reg2type[new_reg.pattern].append(tag.value)
            except IndexError:
                pass

    new_tags = []
    for v in reg2type.keys():
        type_counts = sorted(
            [(the_type, reg2type[v].count(the_type)) for the_type in set(reg2type[v])],
            key=lambda x: (-x[-1], x[0]),
        )
        fav_type = type_counts[0][0]
        regexp = re.compile(v, re.U + re.I * (" " in v))
        for match in regexp.finditer(content):
            lo_tok = word_spans.spans.index([t for t in word_spans if t.lb == match.start()][0])
            hi_tok = word_spans.spans.index([t for t in word_spans if t.ub == match.end()][0]) + 1
            new_tags.append(Tag(fav_type, lo_tok, hi_tok))

    to_remove_tags = []
    for new_tag in new_tags:
        to_remove_tags.extend(
            [
                ann
                for ann in document.annotationset(field)
                if new_tag.lb <= ann.lb and ann.ub <= new_tag.ub and ann.value == new_tag.value
            ]
        )
    for to_remove_tag in to_remove_tags:
        try:
            document.annotationset(field)._annotations.remove(to_remove_tag)
        except ValueError:
            pass

    all_tags = [sent.feature(field) for sent in document.corpus.sentences]
    new_tags.sort(key=lambda x: (x.lb, -x.ub))
    for new_tag in new_tags:
        nth_word = 0
        nth_sent = 0
        sents = document.corpus.sentences
        while nth_word + len(sents[nth_sent]) - 1 < new_tag.lb:
            nth_word += len(sents[nth_sent])
            nth_sent += 1
        start = new_tag.lb - nth_word
        end = new_tag.ub - nth_word
        document.corpus.sentences[nth_sent][start][field] = "B-{}".format(new_tag.value)
        all_tags[nth_sent][start] = "B-{0}".format(new_tag.value)
        for index in range(start + 1, end):
            document.corpus.sentences[nth_sent][index][field] = "I-{0}".format(new_tag.value)
            all_tags[nth_sent][index] = "I-{0}".format(new_tag.value)

    document.add_annotation_from_tags(all_tags, field, field)


def non_overriding_label_consistency(sentence, form2entity, trie, entry, ne_entry):
    length = len(sentence)
    res = sentence.feature(ne_entry)[:]
    tmp = trie
    fst = 0
    lst = -1  # last match found
    cur = 0
    ckey = None  # Current KEY
    tokens = sentence.feature(entry)
    while fst < length - 1:
        cont = True
        while cont and (cur < length):
            ckey = tokens[cur]
            if res[cur] == "O":
                if NUL in tmp:
                    lst = cur
                tmp = tmp.get(ckey, {})
                cont = len(tmp) != 0
                cur += int(cont)
            else:
                cont = False

        if NUL in tmp:
            lst = cur

        if lst != -1:
            form = " ".join([tokens[i] for i in range(fst, lst)])
            appendice = "-{}".format(form2entity[form])
            res[fst] = "B{}".format(appendice)
            for i in range(fst + 1, lst):
                res[i] = "I{}".format(appendice)
            fst = lst
            cur = fst
        else:
            fst += 1
            cur = fst

        tmp = trie
        lst = -1

    if NUL in trie.get(tokens[-1], []) and res[-1] == "O":
        res[-1] = "B-{}".format(form2entity[tokens[-1]])

    return res


def overriding_label_consistency(sentence, form2entity, trie, entry, ne_entry):
    length = len(sentence)
    res = ["O" for _ in range(length)]
    tmp = trie
    fst = 0
    lst = -1  # last match found
    cur = 0
    ckey = None  # Current KEY
    entities = []
    tokens = sentence.feature(entry)
    while fst < length - 1:
        cont = True
        while cont and (cur < length):
            ckey = tokens[cur]
            if res[cur] == "O":
                if NUL in tmp:
                    lst = cur
                tmp = tmp.get(ckey, {})
                cont = len(tmp) != 0
                cur += int(cont)
            else:
                cont = False

        if NUL in tmp:
            lst = cur

        if lst != -1:
            form = " ".join([tokens[i] for i in range(fst, lst)])
            entities.append(Tag(form2entity[form], fst, lst))
            fst = lst
            cur = fst
        else:
            fst += 1
            cur = fst

        tmp = trie
        lst = -1

    if NUL in trie.get(tokens[-1], []):
        entities.append(
            Tag(form2entity[tokens[-1]], length - 1, length)
        )

    gold = chunk_annotation_from_sentence(sentence, ne_entry).annotations

    for i in reversed(range(len(entities))):
        e = entities[i]
        for r in gold:
            if r.lb == e.lb and r.ub == e.ub:
                del entities[i]
                break

    for i in reversed(range(len(gold))):
        r = gold[i]
        for e in entities:
            if r.lb >= e.lb and r.ub <= e.ub:
                del gold[i]
                break

    for r in gold + entities:
        appendice = "-{}".format(r.value)
        res[r.lb] = "B{}".format(appendice)
        for i in range(r.lb + 1, r.ub):
            res[i] = "I{}".format(appendice)

    return res


class LabelConsistencyProcessor(Processor):
    def __init__(
        self,
        field,
        token_field="word",
        label_consistency="overriding",
        **kwargs,
    ):
        super(LabelConsistencyProcessor, self).__init__(**kwargs)
        self._field = field
        self._token_field = token_field

        if label_consistency == "overriding":
            self._feature = overriding_label_consistency
        else:
            self._feature = non_overriding_label_consistency

    def process_document(self, document, abbreviation_resolution=True, **kwargs):
        corpus = document.corpus.sentences
        field = self._field
        token_field = self._token_field

        entities = {}
        counts = {}
        for p in corpus:
            G = chunk_annotation_from_sentence(p, column=field)
            for entity in G:
                id = entity.value
                form = " ".join(p.feature(token_field)[entity.lb: entity.ub])
                if form not in counts:
                    counts[form] = {}
                if id not in counts[form]:
                    counts[form][id] = 0
                counts[form][id] += 1

        for form, count in counts.items():
            if len(count) == 1:
                entities[form] = list(count.keys())[0]
            else:
                best = sorted(count.keys(), key=lambda x: -count[x])[0]
                entities[form] = best

        value = Trie()
        for entry in entities.keys():
            entry = entry.strip()
            if entry:
                value.add(entry.split())

        for p in corpus:
            p.add(self._feature(p, entities, value.data, token_field, field), field)

        tags = [sentence.feature(field) for sentence in corpus]
        document.add_annotation_from_tags(tags, field, field)

        if abbreviation_resolution:
            detect_abbreviations(document, field)


class MapAnnotationsProcessor(Processor):
    def __init__(self, mapping, annotation_name, **kwargs):
        super(MapAnnotationsProcessor, self).__init__(**kwargs)

        if isinstance(mapping, str):
            with open(mapping, 'r', encoding="utf-8") as input_stream:
                self._mapping = compile_map(input_stream)
        else:
            self._mapping = mapping

        self._annotation_name = annotation_name

    def process_document(self, document, **kwargs):
        """Updates a document with various segmentations and creates
        an sem.corpus (CoNLL-formatted data) using field argument as index.

        Parameters
        ----------
        document : sem.storage.Document
            the input data. It is a document with only a content
        """

        start = time.time()

        ref_annotation = document.annotationset(self._annotation_name)
        ref_annotations = ref_annotation.annotations
        new_annotations = [
            Tag(self._mapping.get(annotation.value, annotation.value), annotation.lb, annotation.ub)
            for annotation in ref_annotations
            if self._mapping.get(annotation.value, None) != ""
        ]

        document.add_annotationset(
            AnnotationSet(
                self._annotation_name,
                reference=ref_annotation.reference,
                annotations=new_annotations,
            )
        )

        laps = time.time() - start
        sem.logger.info("in %s", timedelta(seconds=laps))


class ExportProcessor(Processor):
    def __init__(
        self,
        exporter,
        lang="fr",
        lang_style="default.css",
        pos_column=None,
        chunk_column=None,
        ner_column=None,
        **kwargs,
    ):
        super(ExportProcessor, self).__init__(**kwargs)

        self._lang = lang
        self._lang_style = lang_style
        self._pos_column = pos_column
        self._chunk_column = chunk_column
        self._ner_column = ner_column
        if isinstance(exporter, str):
            sem.logger.info("getting exporter {0}".format(exporter))
            Exporter = get_exporter(exporter)
            self._exporter = Exporter(lang=self._lang, lang_style=self._lang_style)
        else:
            sem.logger.info("using loaded exporter")
            self._exporter = exporter

    def process_document(self, document, outfile=sys.stdout, output_encoding="utf-8", **kwargs):
        start = time.time()

        sem.logger.debug("setting name/column couples for exportation")

        pos_column = self._pos_column
        chunk_column = self._chunk_column
        ner_column = self._ner_column

        couples = {}
        if "word" in document.corpus.fields:
            couples["token"] = "word"
        elif "token" in document.corpus.fields:
            couples["token"] = "token"

        if pos_column:
            couples["pos"] = pos_column
            sem.logger.debug("POS column is {0}".format(pos_column))
        if chunk_column:
            couples["chunking"] = chunk_column
            sem.logger.debug("chunking column is {0}".format(chunk_column))
        if ner_column:
            couples["ner"] = ner_column
            sem.logger.debug("NER column is {0}".format(ner_column))

        sem.logger.debug("exporting document to {0} format".format(self._exporter.extension))

        self._exporter.document_to_file(document, couples, outfile, encoding=output_encoding)

        laps = time.time() - start
        sem.logger.info("done in %s", timedelta(seconds=laps))


class PymorphyProcessor(Processor):
    def __init__(self, token_field="word", *args, **kwargs):
        super(PymorphyProcessor, self).__init__(**kwargs)

        self._token_field = token_field
        self._morph = pymorphy2.MorphAnalyzer()

    def process_document(self, document, **kwargs):
        start = time.time()
        annotations = AnnotationSet("POS", reference=document.segmentation("tokens"))
        current = 0
        for sentence in document.corpus:
            lemma = []
            pos = []
            for token in sentence.feature(self._token_field):
                analyzed = self._morph.parse(token)
                lemma.append(analyzed[0].normal_form)
                pos.append(str(analyzed[0].tag.POS or analyzed[0].tag).split(',')[0])
                annotations.append(Tag(pos[-1], current, current+1))
                current += 1
            sentence.add(lemma[:], "lemma")
            sentence.add(pos[:], "POS")
        document.add_annotationset(annotations)
        laps = time.time() - start
        sem.logger.info(u'done in %fs' % laps)


__name2class = {
    "segmentation": SegmentationProcessor,
    "enrich": EnrichProcessor,
    "wapiti_label": WapitiLabelProcessor,
    "clean": CleanProcessor,
    "label_consistency": LabelConsistencyProcessor,
    "map_annotations": MapAnnotationsProcessor,
}


def build_processor(name, additional_names=None, **kwargs):
    name2class = __name2class
    name2class.update(additional_names or {})
    return name2class[name](**kwargs)



# -*- coding: utf-8 -*-

"""
file: tagger.py

Description: performs a sequence of operations in a pipeline.

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse
import os
import shutil
import multiprocessing
import time
import pathlib
from datetime import timedelta
from functools import partial

import configparser

import sem
import sem.logger
import sem.importers
import sem.pipelines

if sem.ON_WINDOWS:
    sem.logger.warning(
        "multiprocessing not handled on Windows. Documents will be processed sequentially."
    )

__pipeline = None


def process(document, exporter, output_directory, couples, encoding, lang_style):
    """
    The function used to allow multiprocessing of documents.
    Note that multiprocessing will only work on Linux.
    The function is written to work sequentially on Windows to avoid dupe.
    """
    __pipeline.process_document(document)

    if exporter is not None:
        name = document.escaped_name()
        if "html" in exporter.extension:
            shutil.copy(sem.SEM_RESOURCE_DIR / "css" / "tabs.css", output_directory)
            shutil.copy(
                sem.SEM_RESOURCE_DIR / "css" / exporter._lang / lang_style, output_directory
            )

        shortname = pathlib.Path(name).stem
        out_path = output_directory / "{0}.{1}".format(shortname, exporter.extension)
        if exporter.extension == "ann":
            filename = shortname + ".txt"
            with open(output_directory / filename, "w", encoding=encoding) as output_stream:
                output_stream.write(document.content)
        exporter.document_to_file(document, couples, out_path, encoding=encoding)

    return document


def get_option(cfg, section, option, default=None):
    try:
        return cfg.get(section, option)
    except (configparser.NoSectionError, configparser.NoOptionError):
        return default


def get_section(cfg, section):
    try:
        return dict(cfg.items(section))
    except configparser.NoSectionError:
        return {}


def main(argv=None):
    args = parser.parse_args(argv)
    tagger(**vars(args))


def tagger(
    workflow,
    infiles,
    output_directory=".",
    force_format="default",
    n_procs=1,
    pipeline=None,
    options=None,
    exporter=None,
    couples=None
):
    """Return a document after it passed through a pipeline.

    Parameters
    ----------
    workflowfile : str
        the file containing the pipeline and global options
    infiles : list
        the input for the upcoming pipe. Its base value is the list of files to
        treat, it can be either "plain text" or CoNNL-formatted file.
    directory : str
        the directory where every file will be outputted.
    n_procs : int
        the number of processors to use to process documents.
        f n_procs < 0, it will be set to 1, if n_procs == 0 it will be set
        to cpu_count, and n_procs otherwise.
        If n_procs is greater than the number of documents, it will be
        adjusted.
    """

    start = time.time()

    global __pipeline

    output_directory = pathlib.Path(output_directory)

    if pipeline is None or options is None:
        workflow = pathlib.Path(workflow)
        pipeline, options, exporter, couples = sem.pipelines.load_workflow(workflow, force_format)
    __pipeline = pipeline

    if get_option(options, "log", "log_file") is not None:
        sem.logger.addHandler(sem.logger.file_handler(get_option(options, "log", "log_file")))
    sem.logger.setLevel(get_option(options, "log", "log_level", "WARNING"))

    if not output_directory.exists():
        os.makedirs(output_directory)

    oenc = get_option(options, "encoding", "output_encoding", "utf-8")

    file_format = get_option(options, "file", "format", "guess")
    opts = get_section(options, "file")
    opts.update(get_section(options, "encoding"))
    if file_format == "conll":
        opts["fields"] = opts["fields"].split(",")
        opts["taggings"] = [tagging for tagging in opts.get("taggings", "").split(",") if tagging]
        opts["chunkings"] = [
            chunking for chunking in opts.get("chunkings", "").split(",") if chunking
        ]

    documents = sem.importers.documents_from_list(infiles, file_format, **opts)

    n_procs = int(n_procs)
    if n_procs == 0:
        n_procs = multiprocessing.cpu_count()
        sem.logger.info("no processors given, using %s", n_procs)
    else:
        n_procs = min(max(n_procs, 1), multiprocessing.cpu_count())
    if n_procs > len(documents):
        n_procs = len(documents)

    do_process = partial(
        process,
        exporter=exporter,
        output_directory=output_directory,
        couples=couples,
        encoding=oenc,
        lang_style=get_option(options, "export", "lang_style", "default.css"),
    )
    if sem.ON_WINDOWS or n_procs == 1:
        for document in documents:
            do_process(document)
    else:
        pool = multiprocessing.Pool(processes=n_procs)
        dpp = 1 if n_procs < len(documents) * 2 else 2  # documents per processor
        beg = 0
        batch_size = dpp * n_procs
        while beg <= len(documents):
            documents[beg: beg + batch_size] = pool.map(
                do_process, documents[beg: beg + batch_size]
            )
            beg += batch_size
        pool.terminate()

    laps = time.time() - start
    sem.logger.info("done in %s", timedelta(seconds=laps))

    return documents


parser = argparse.ArgumentParser(
    "Performs various operations given in a workflow configuration file that defines a pipeline."
)

parser.add_argument(
    "workflow",
    help="The workflow configuration file."
    " Defines at least the pipeline and may provide some options.",
)
parser.add_argument("infiles", nargs="+", help="The input file(s) for the tagger.")
parser.add_argument(
    "-o",
    "--output-directory",
    dest="output_directory",
    default=".",
    help='The output directory (default: "%(default)s").',
)
parser.add_argument(
    "-f",
    "--force-format",
    default="default",
    help='Force the output format given in "workflow" (default: "%(default)s").',
)
parser.add_argument(
    "-p",
    "--processors",
    dest="n_procs",
    type=int,
    default=1,
    help='The number of processors to use (default: "%(default)s").',
)



# -*- coding: utf-8 -*-

"""
file: enrich.py

Description: this program is used to enrich a CoNLL-formatted file with
various features.

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse
import time
from datetime import timedelta

from sem.processors import EnrichProcessor

import sem.util
import sem.logger
from sem.importers import conll_file


def main(argv=None):
    enrich(**vars(parser.parse_args(argv)))


def enrich(
    infile,
    infofile,
    outfile,
    mode="train",
    ienc=None,
    oenc=None,
    enc="utf-8",
    log_level="WARNING",
    log_file=None
):
    """
    Takes a CoNLL-formatted file and write another CoNLL-formatted file
    with additional features in it.

    Parameters
    ----------
    infile : str
        the CoNLL-formatted input file.
    infofile : str
        the XML file containing the different features.
    mode : str
        the mode to use for infofile. Some inputs may only be present in
        a particular mode. For example, the output tag is only available
        in "train" mode.
    log_level : str or int
        the logging level.
    log_file : str
        if not None, the file to log to (does not remove command-line
        logging).
    """

    start = time.time()

    if log_file is not None:
        sem.logger.addHandler(sem.logger.file_handler(log_file))
    sem.logger.setLevel(log_level)
    sem.logger.info('parsing enrichment file "%s"', infofile)

    processor = EnrichProcessor(path=infofile, mode=mode)

    sem.logger.debug('enriching file "%s"', infile)

    bentries = [entry.name for entry in processor.bentries]
    aentries = [entry.name for entry in processor.aentries]
    document = conll_file(
        infile, bentries + aentries, (bentries + aentries)[0], encoding=ienc or enc
    )

    processor.process_document(document)
    fields = processor.fields()
    with open(outfile, "w", encoding=oenc or enc) as output_stream:
        for sentence in document.corpus:
            output_stream.write(sentence.conll(fields))
            output_stream.write("\n\n")

    laps = time.time() - start
    sem.logger.info("done in %s", timedelta(seconds=laps))


parser = argparse.ArgumentParser(
    "Adds information to a file using and XML-styled configuration file."
)

parser.add_argument("infile", help="The input file (CoNLL format)")
parser.add_argument("infofile", help="The information file (XML format)")
parser.add_argument("outfile", help="The output file (CoNLL format)")
parser.add_argument(
    "-m",
    "--mode",
    dest="mode",
    default="train",
    choices=("train", "label", "annotate", "annotation"),
    help="The mode for enrichment. May make entries vary (default: %(default)s)",
)
parser.add_argument("--input-encoding", dest="ienc", help="Encoding of the input (default: UTF-8)")
parser.add_argument("--output-encoding", dest="oenc", help="Encoding of the input (default: UTF-8)")
parser.add_argument(
    "--encoding",
    dest="enc",
    default="UTF-8",
    help="Encoding of both the input and the output (default: UTF-8)",
)
parser.add_argument(
    "-l",
    "--log",
    dest="log_level",
    choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
    default="WARNING",
    help="Increase log level (default: critical)",
)
parser.add_argument("--log-file", dest="log_file", help="The name of the log file")



# -*- coding: utf-8 -*-

"""
file: features.py

Description: SEM features for Corpus type.

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import re
import pathlib
import functools

from sem.constants import NUL
from sem.storage import (
    AnnotationSet,
    compile_token,
    compile_multiword,
    get_top_level,
    chunks_to_annotation
)


def get(x):
    """The method to automatically get an element in a list with a shift.
    This is useful for features like "previous word starts with uppercase".

    Parameters
    ----------
    x : int
        the shift to apply to an index

    Returns
    -------
    function
        A function that will get the element of a list with a shift of x.
        When no element is found, the method will return empty string.
    """
    def get_sent(lst, index):
        if 0 <= index + x < len(lst):
            return lst[index+x]
        return ""
    return get_sent


def sentence_wrapper(f, x, y):
    """A wrapper for methods that work on a single token (like an str) to make them work
    on a Sentence (list of elements).

    Parameters
    ----------
    f : function
        the function to apply on a sentence.
    x : int
        the shift in position in sentence.
    y : str
        the field to apply f on.

    Returns
    -------
    function
        a mapping function that will map f on a Sentence object.
    """
    getter = get(x)

    def do_map(s):
        features = s.feature(y)
        return [f(getter(features, i)) for i in range(len(s))]
    return do_map


def limited_compose(funs):
    """Minimalistic function composition by messing with `functools.reduce`.
    This method is limited as it requires a single argument, where valid call
    could be no arguments or many. This means that some valid calls will raise
    an exception (the message will also be very obscure if do not know what is
    going on).
    This function is intended to be used as a mean to model sequence features
    is SEM, where multiple functions are called one after the other.

    For real composition, use `funcy.funcs.compose` (https://github.com/Suor/funcy).

    Parameters
    ----------
    funs : list[function]
        The list of one argument functions to compose.

    Returns
    -------
    function
        The composition of functions given in argument.

    Raises
    ------
    ValueError
        There are less than 2 functions to compose.
    """
    if len(funs) < 2:
        raise ValueError("Cannot compose less than 2 functions")
    # `functools.reduce` works using an accumulator that will hold
    # the current value. This accumulator is `x` in the lambda expression,
    # y is the current function in the iterator.
    # We can see why we *need* an argument: if no arguments are given, reduce
    # will assume the first element of the list to be the initializer, meaning
    # the first element will be a function.
    return functools.partial(functools.reduce, lambda x, y: y(x), funs)


def is_upper_at(token, index):
    return token[index].isupper()


def bos(s):
    return [i == 0 for i in range(len(s))]


def eos(s):
    return [i == len(s)-1 for i in range(len(s))]


def matches(s, regexp, y):
    return [regexp.search(token) is not None for token in s.feature(y)]


def subsequence(s, regexp, y, default="#"):
    lst = [regexp.search(token) for token in s.feature(y)]
    return [item.group() if item else default for item in lst]


def token(s, regexp, y, default="#"):
    return [(token if regexp.search(token) else default) for token in s.feature(y)]


def check_some(s, features):
    return [any(feats) for feats in zip(*[f(s) for f in features])]


def check_all(s, features):
    return [all(feats) for feats in zip(*[f(s) for f in features])]


def check_none(s, features):
    return [not any(feats) for feats in zip(*[f(s) for f in features])]


def substitute(string, pattern, repl):
    """Just an interface to make re.sub work with functools.partial"""
    return re.sub(pattern, repl, string)


def boolean_format(fmt):
    if fmt.lower() in ("bo", "b/o"):
        return bo_fmt
    elif fmt.lower() in ("zo", "z/o", "01", "0/1"):
        return zo_fmt
    else:
        raise ValueError(f'unknown boolean feature format: "{fmt}"')


def bo_fmt(value, appendice):
    """Begin/Out format."""
    return ('B{}'.format(appendice) if value else 'O')


def zo_fmt(value, appendice):
    """Zero/One format."""
    return int(value)


def not_op(s, f, fmt=zo_fmt):
    return [fmt(not item, '') for item in f(s)]


def and_op(s, f1, f2, fmt=zo_fmt):
    return [fmt(left and right, '') for left, right in zip(f1(s), f2(s))]


def token_gazetteer(s, gazetteer, y, appendice='', fmt=bo_fmt):
    return [fmt(token in gazetteer, appendice) for token in s.feature(y)]


def multiword_dictionary(s, trie, y, appendice=''):
    length = len(s)
    res = ['O' for _ in range(length)]
    tmp = trie.data
    fst = 0
    lst = -1  # last match found
    cur = 0
    ckey = None  # Current KEY
    tokens = s.feature(y)
    while fst < length - 1:
        cont = True
        while cont and (cur < length):
            ckey = tokens[cur]
            if NUL in tmp:
                lst = cur
            tmp = tmp.get(ckey, {})
            cont = len(tmp) != 0
            cur += int(cont)
        if NUL in tmp:
            lst = cur
        if lst != -1:
            res[fst] = "B{}".format(appendice)
            for i in range(fst + 1, lst):
                res[i] = "I{}".format(appendice)
            fst = lst
            cur = fst
        else:
            fst += 1
            cur = fst
        tmp = trie.data
        lst = -1
    if NUL in trie.data.get(tokens[-1], []):
        res[-1] = "B{}".format(appendice)
    return res


def directory_feature(s, features):
    data = ["O" for _ in range(len(s))]
    annotation = AnnotationSet("")
    for feat in features:
        for annot in chunks_to_annotation(feat(s)):
            annotation.add(annot)
    tags = get_top_level(annotation)
    for tag in tags:
        data[tag.lb] = "B-{}".format(tag.value)
        for index in range(tag.lb + 1, tag.ub):
            data[index] = "I-{}".format(tag.value)
    return data


def apply_mapping(s, mapping, y, default="O"):
    return [mapping.get(item, default) or default for item in s.feature(y)]


def fill(s, condition, entry, filler):
    l1 = s.feature(entry)
    l2 = s.feature(filler)
    cond = condition(s)
    return [(filling if c else data) for (data, filling, c) in zip(l1, l2, cond)]


def search_forward(s, condition, entry, default='#'):
    tokens = s.feature(entry)
    matches = condition(s)
    lst = []
    to_add = default
    for i, item in reversed(list(enumerate(matches))):
        lst.append(to_add)
        if item:
            to_add = tokens[i]
    return lst[::-1]


def search_backward(s, condition, entry, default='#'):
    tokens = s.feature(entry)
    matches = condition(s)
    lst = []
    to_add = default
    for i, item in enumerate(matches):
        lst.append(to_add)
        if item:
            to_add = tokens[i]
    return lst


def xml2feat(xml, default_entry="word", default_shift=0, path=None):
    attrib = xml.attrib
    action = attrib.get("action")
    y = attrib.get("entry", default_entry)
    x = attrib.get("shift", default_shift)

    if xml.tag == "boolean":
        fmt = boolean_format(attrib.get("format", "0/1"))
        if action.lower() == "and":
            left, right = list(xml)
            return functools.partial(and_op, f1=xml2feat(left), f2=xml2feat(right), fmt=fmt)
        elif action.lower() == "not":
            return functools.partial(not_op, f=xml2feat(list(xml)[0]), fmt=fmt)

    if xml.tag == "string":
        if action.lower() == "equal":
            casing = attrib.get("casing", "s").lower()
            if casing in ("s", "sensitive"):
                return sentence_wrapper(functools.partial(str.__eq__, xml.text), x=x, y=y)
            elif casing in ("i", "insensitive"):
                return sentence_wrapper(
                    limited_compose([str.lower, functools.partial(str.__eq__, xml.text.lower())]),
                    x=x,
                    y=y
                )
        else:
            raise ValueError("Invalid string action: {}".format(action))

    elif xml.tag == "nullary":
        if action.lower() == "lower":
            return sentence_wrapper(str.lower, x=x, y=y)
        elif action.lower() == "bos":
            return bos
        else:
            raise ValueError("Invalid nullary action: {}".format(action))

    elif xml.tag == "unary":
        if action.lower() == "isupper":
            return sentence_wrapper(
                functools.partial(is_upper_at, index=int(xml.text)), x=x, y=y
            )
        else:
            return ValueError("Invalid unaty action: {}".format(action))

    elif xml.tag == "directory":
        children = list(xml)
        if children:
            features = []
            for child in reversed(children):
                features.append(xml2feat(child, path=path))
            return functools.partial(directory_feature, features=features)

    elif xml.tag == "dictionary":
        if "path" in attrib:
            if path:
                the_path = pathlib.Path(path).parent / xml.attrib["path"]
            else:
                the_path = xml.attrib["path"]
            with open(the_path) as input_stream:
                entries = input_stream.readlines()
            appendice = "-{}".format(pathlib.Path(xml.attrib["path"]).stem)
        elif xml.text:
            entries = xml.text.split("\n")
        else:
            raise ValueError("Dictionary feature should have either a path attribute or text.")

        action = attrib["action"].lower()
        entry = attrib.get("entry", "word")
        if action == "token":
            appendice = attrib.get("appendice", appendice)
            fmt = boolean_format(attrib.get("format", "B/O"))
            data = compile_token(entries)
            return functools.partial(
                token_gazetteer, gazetteer=data, y=entry, appendice=appendice, fmt=fmt
            )
        elif action == "multiword":
            appendice = attrib.get("appendice", appendice)
            data = compile_multiword(entries)
            return functools.partial(multiword_dictionary, trie=data, y=entry, appendice=appendice)
        else:
            raise ValueError("Invalid action for dictionary: {}".format(action))

    elif xml.tag == "fill":
        entry = attrib.pop("entry")
        filler_entry = attrib.pop("filler-entry", default_entry)
        condition = xml2feat(list(xml)[0])
        return functools.partial(fill, condition=condition, entry=entry, filler=filler_entry)

    elif xml.tag == "find":
        action = attrib["action"].lower()
        default = attrib.get("default", "#")
        if action == "forward":
            condition = xml2feat(list(xml)[0])
            return functools.partial(search_forward, condition=condition, entry=y, default=default)
        elif action == "backward":
            condition = xml2feat(list(xml)[0])
            return functools.partial(search_backward, condition=condition, entry=y, default=default)
        else:
            raise ValueError("Invalid action for find: {}".format(action))

    elif xml.tag == "regexp":
        flags = re.U + re.M + (re.I * int("i" == xml.attrib.get("casing", "s")))
        if attrib["action"].lower() == "check":
            return functools.partial(matches, regexp=re.compile(xml.text, flags=flags), y=y)
        raise RuntimeError

    else:
        raise ValueError("Invalid feature kind: {}".format(xml.tag))



"""
file: gui_components.py

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import re

import tkinter
import tkinter.ttk
import tkinter.filedialog
import tkinter.messagebox

import pathlib
import os
import multiprocessing
import time
import shutil
import configparser

import sem
import sem.exporters
import sem.importers
import sem.logger
import sem.pipelines

from sem.constants import NUL
from sem.storage import SEMCorpus, str2filter, str2docfilter, Tag, Trie
from sem.processors import EnrichProcessor, WapitiLabelProcessor
from sem.modules.tagger import tagger

from wapiti.api import Model as WapitiModel


SPARE_COLORS_DEFAULT = [
    {'background': '#FFCCCC', 'foreground': '#FF0000'},
    {'background': '#CCEEEE', 'foreground': '#008888'},
    {'background': '#CCCCFF', 'foreground': '#0000FF'},
    {'background': '#DDFFDD', 'foreground': '#008800'},
    {'foreground': '#6c4c45', 'background': '#e6dad7'},
    {'foreground': '#813058', 'background': '#eecfde'},
    {'foreground': '#8a570d', 'background': '#f4c888'},
    {'foreground': '#1461a1', 'background': '#cce5f9'},
    {'foreground': '#601194', 'background': '#d7a8f6'},
    {'foreground': '#254084', 'background': '#bccaed'},
    {'foreground': '#a22800', 'background': '#ffb299'},
    {'foreground': '#729413', 'background': '#e3f5af'},
    {'foreground': '#0a9b47', 'background': '#a3fac8'},
    {'foreground': '#275a5f', 'background': '#85c6cc'},
    {'foreground': '#886c11', 'background': '#f1da91'},
    {'foreground': '#426722', 'background': '#aad684'},
    {'background': '#C9B297', 'foreground': '#5C4830'},
    {'background': '#C8A9DC', 'foreground': '#542D6E'},
    {'foreground': '#79a602', 'background': '#e7fea8'},
    {'foreground': '#454331', 'background': '#a7a383'},
    {'foreground': '#625e2d', 'background': '#d0cb99'},
    {'foreground': '#4b3054', 'background': '#b28fbf'},
    {'foreground': '#374251', 'background': '#9ca9bc'},
    {'background': '#CCCCCC', 'foreground': '#000000'},
]


def fill_with(t, value):
    def fill_rec(t1):
        keys = [key for key in t1 if key != NUL]
        for key in keys:
            fill_rec(t1[key])
        t1[NUL] = value[:]

    fill_rec(t.data)


def find_potential_separator(target):
    regex = re.compile(r"(\W+)", re.U)
    found = list(set(regex.findall(target)))
    if len(found) == 1:
        return found[0]
    return None


def find_occurrences(target, content, whole_word=True):
    target = target.strip()
    if whole_word:
        pattern = (
            (r"\b" if target[0].isalnum() else r"((?<=\s)|(?<=^))")
            + re.escape(target)
            + (r"\b" if target[-1].isalnum() else r"(?=\s|$)")
        )
    else:
        pattern = re.escape(target)
    regex = re.compile(pattern, re.U + re.M)
    for match in regex.finditer(content):
        yield match


def random_color():
    import random
    import colorsys

    red = (random.randrange(0, 256) + 256) / 2.0
    green = (random.randrange(0, 256) + 256) / 2.0
    blue = (random.randrange(0, 256) + 256) / 2.0

    def to_hex(i):
        hx = hex(int(i))[2:]
        if len(hx) == 1:
            hx = "0{}".format(hx)
        return hx

    def to_color(r, g, b):
        cs = [to_hex(c) for c in [r, g, b]]
        return "#{cs[0]}{cs[1]}{cs[2]}".format(cs=cs)

    def darker(r, g, b):
        h, l, s = colorsys.rgb_to_hls(r / 256.0, g / 256.0, b / 256.0)
        cs = [to_hex(256.0 * c) for c in colorsys.hls_to_rgb(h, l / 2.5, s)]
        return "#{cs[0]}{cs[1]}{cs[2]}".format(cs=cs)

    return {"background": to_color(red, green, blue), "foreground": darker(red, green, blue)}


class Adder:
    l2t = {}
    t2l = {}

    @classmethod
    def label2type(cls, label, level, default=None):
        return cls.l2t.get(level, {}).get(label, default)

    @classmethod
    def type2label(cls, type, level, default=None):
        return cls.t2l.get(level, {}).get(type, default)

    @classmethod
    def clear(cls):
        cls.l2t.clear()
        cls.t2l.clear()

    def __init__(self, frame, the_type, available, level=0):
        if len(available[level]) == 0:
            raise ValueError("No more available shortcuts!")
        self.frame = frame
        self.type = the_type[:]
        self.level = level
        self.label = the_type.lower()
        self.shortcut = None
        found = False
        if self.level == 0:
            if len(self.frame.spare_colors) > 0:
                self.color = self.frame.spare_colors.pop()
            else:
                self.color = random_color()
            self.frame.text.tag_configure(self.type, **self.color)
        for i in range(len(self.label)):
            target = self.label[i].lower()
            found = target in available[level]
            if found:
                available[level].remove(target)
                self.shortcut = target
                self.label = self.label + " [{0} or Shift+{0}]".format(self.label[i])
                break
        if not found and len(available[level]) > 0:
            char = available[level][0]
            available[level].remove(char)
            self.shortcut = char
            self.label += " [{0} or Shift+{0}]".format(char)
        if self.level not in Adder.l2t:
            Adder.l2t[self.level] = {}
            Adder.t2l[self.level] = {}
        Adder.l2t[self.level][self.label] = self.type
        Adder.t2l[self.level][self.type] = self.label

    def add(self, event, remove_focus=False):
        if self.frame.current_selection is not None:
            f_cs = self.frame.current_selection
            tag = Tag(self.type, f_cs.lb, f_cs.ub)
            first = self.frame.charindex2position(f_cs.lb)
            last = self.frame.charindex2position(f_cs.ub)
            if (
                tag in self.frame.current_annotations
                and self.frame.current_type_hierarchy_level == 0
            ):
                return
        else:
            first = "sel.first"
            last = "sel.last"
        self.frame.wish_to_add = [self.type, first, last]
        self.frame.add_annotationset(None, remove_focus)

    def add_all(self, event):
        if self.frame.current_selection is not None:
            start = self.frame.charindex2position(self.frame.current_selection.lb)
            end = self.frame.charindex2position(self.frame.current_selection.ub)
        else:
            start, end = ("sel.first", "sel.last")
        try:
            target = re.escape(self.frame.text.get(start, end).strip())
            pattern = (
                (r"\b" if target[0].isalnum() else r"((?<=\s)|(?<=^))")
                + target
                + (r"\b" if target[-1].isalnum() else r"(?=\s|$)")
            )
            regex = re.compile(pattern, re.U + re.M)
            for match in regex.finditer(self.frame.doc.content):
                cur_start = self.frame.charindex2position(match.start())
                cur_end = self.frame.charindex2position(match.end())
                if Tag(self.type, match.start(), match.end()) not in self.frame.current_annotations:
                    self.frame.wish_to_add = [self.type, cur_start, cur_end]
                    self.frame.add_annotationset(None, remove_focus=False)
        except tkinter.TclError:
            raise
        self.frame.type_combos[self.level].current(0)
        self.frame.wish_to_add = None
        self.frame.current_selection = None
        self.frame.current_type_hierarchy_level = 0
        self.frame.update_level()
        self.frame.text.tag_remove("BOLD", "1.0", "end")


def makemap(tagset, colors):
    dct = {}
    for tag in tagset:
        val = tag.split(".", 1)[0]
        color = (colors.pop(0) if colors else random_color())
        dct[val] = color
    return dct


class Adder2:
    def __init__(self, tagset, levels, shortcut_trie, color=None):
        self.tagset = tagset
        self.levels = levels
        self.shortcut_trie = shortcut_trie
        self.current_annotation = None
        self.current_hierarchy_level = 0
        self.color = color or makemap(self.tagset, SPARE_COLORS_DEFAULT[:])

    def max_depth(self):
        return max([len(lvl) for lvl in self.levels])

    def up_one_level(self):
        if self.current_annotation is None:
            self.current_hierarchy_level = 0
        else:
            self.current_hierarchy_level += 1
            if self.current_hierarchy_level >= self.max_depth():
                self.current_hierarchy_level = 0

    def down_one_level(self):
        if self.current_annotation is None:
            self.current_hierarchy_level = 0
        else:
            self.current_hierarchy_level -= 1
            if self.current_hierarchy_level < 0:
                self.current_hierarchy_level = self.max_depth() - 1

    def type_from_letter(self, letter):
        if (
            self.current_annotation is not None
            and len(self.current_annotation.levels) < self.current_hierarchy_level
        ):
            return None
        path = (
            self.current_annotation.levels[: self.current_hierarchy_level]
            if self.current_annotation
            else []
        )
        sub = self.shortcut_trie.goto(path)
        for key, val in sub.items():
            if key != NUL and val[NUL] == letter:
                return key


def from_tagset(tagset):
    levels = [tag.split(".") for tag in tagset]
    chars = list("abcdefghijklmnopqrstuvwxyz") + [
        "F1",
        "F2",
        "F3",
        "F4",
        "F5",
        "F6",
        "F7",
        "F8",
        "F9",
        "F10",
        "F11",
        "F12",
        "*",
    ]
    trie = Trie()
    for level in levels:
        trie.add(level)
    fill_with(trie, chars)
    shortcut_trie = Trie()
    for level in levels:
        hierarchy = []
        for depth, sublevel in enumerate(level):
            if hierarchy + [sublevel] in shortcut_trie:
                hierarchy.append(sublevel)
                continue
            sub = trie.goto(hierarchy)
            available = sub[NUL]
            for c in sublevel.lower():
                found = c in available
                if found:
                    available.remove(c)
                    break
            if not found:
                c = available[0]
                available.remove(c)
            hierarchy.append(sublevel)
            shortcut_trie.add_with_value(hierarchy, c)
    return Adder2(tagset, levels, shortcut_trie)


class SemTkLangSelector(tkinter.ttk.Frame):
    def __init__(self, root, resource_dir):
        tkinter.ttk.Frame.__init__(self, root)

        self.resource_dir = resource_dir
        self.observers = []
        self.items = [resource.name for resource in self.resource_dir.glob("*")]

        self.cur_lang = tkinter.StringVar()
        self.select_lang_label = tkinter.ttk.Label(root, text="select language:")
        self.langs = tkinter.ttk.Combobox(root, textvariable=self.cur_lang)

        self.langs["values"] = self.items

        self.langs.current(0)
        for i, item in enumerate(self.items):
            if item == "fr":
                self.langs.current(i)

        self.langs.bind("<<ComboboxSelected>>", self.select_lang)

    def pack(self):
        self.select_lang_label.pack()
        self.langs.pack()

    def grid(self, row=0, column=0):
        x = row
        y = column
        self.select_lang_label.grid(row=x, column=y)
        x += 1
        self.langs.grid(row=x, column=y)
        x += 1
        return (x, y)

    def lang(self):
        return self.cur_lang.get()

    def register(self, observer):
        self.observers.append(observer)
        observer.set_lang(self.lang())

    def select_lang(self, event):
        for observer in self.observers:
            observer.set_lang(self.lang())


class SemTkResourceSelector(tkinter.ttk.Frame):
    def __init__(self, root, resource_dir, filter=lambda x: True):
        tkinter.ttk.Frame.__init__(self, root)

        self.resource_dir = resource_dir
        self._filter = filter
        self._lang = None
        self._items = []
        self.select_resourse_label = tkinter.ttk.Label(
            root, text=f"select {self.resource_dir.name}:"
        )
        self.resources = tkinter.Listbox(root)

    def pack(self):
        self.select_resourse_label.pack()
        self.resources.pack()

    def grid(self, row=0, column=0):
        x = row
        y = column
        self.select_resourse_label.grid(row=x, column=y)
        x += 1
        self.resources.grid(row=x, column=y)
        x += 1
        return (x, y)

    def resource(self):
        resrc = self.resources.get(tkinter.ACTIVE)
        return self.resource_dir / self.lang() / resrc or None

    def lang(self):
        return self._lang

    def set_lang(self, language):
        self._lang = language
        self.items = [
            path.name for path in (self.resource_dir / self._lang).glob("*") if self._filter(path)
        ]
        self.items.sort(key=lambda x: x.lower())
        self.resources["width"] = max(
            max(len(item) for item in self.items) + 1, self.resources["width"]
        )
        self.resources["height"] = len(self.items)

        self.resources.delete(0, tkinter.END)
        for item in self.items:
            self.resources.insert(tkinter.END, item)


class SemTkFileSelector(tkinter.ttk.Frame):
    def __init__(self, root, main_window, button_opt):
        tkinter.ttk.Frame.__init__(self, root)

        self.root = root
        self.main_window = main_window
        self.current_files = None
        self.button_opt = button_opt

        # define options for opening or saving a file
        self.file_opt = options = {}
        options["defaultextension"] = ".txt"
        options["filetypes"] = [("all files", ".*"), ("text files", ".txt")]
        options["initialdir"] = pathlib.Path("~").expanduser()
        options["parent"] = root
        options["title"] = "Select files to annotate."

        self.file_selector_button = tkinter.ttk.Button(
            self.root, text="select file(s)", command=self.filenames
        )
        self.label = tkinter.ttk.Label(self.root, text="selected file(s):")
        self.fa_search = tkinter.PhotoImage(
            file=self.main_window.resource_dir / "images" / "fa_search_24_24.gif"
        )
        self.file_selector_button.config(image=self.fa_search, compound=tkinter.LEFT)

        self.scrollbar = tkinter.ttk.Scrollbar(self.root)
        self.selected_files = tkinter.Listbox(self.root, yscrollcommand=self.scrollbar.set)
        self.scrollbar.config(command=self.selected_files.yview)

    def pack(self):
        self.file_selector_button.pack(**self.button_opt)
        self.label.pack()
        self.scrollbar.pack(side=tkinter.RIGHT, fill=tkinter.Y)
        self.selected_files.pack(fill=tkinter.BOTH, expand=True)

    def grid(self, row=0, column=0):
        """
        TODO: to be tested
        """
        x = row
        y = column
        self.file_selector_button.grid(row=x, column=y, **self.button_opt)
        x += 1
        self.label.grid(row=x, column=y)
        x += 1
        self.scrollbar.pack(side=tkinter.RIGHT, fill=tkinter.Y)
        self.selected_files.grid(row=x, column=y, fill=tkinter.BOTH, expand=True)
        x += 1
        return (x, y)

    def filenames(self, event=None):
        self.current_files = list(tkinter.filedialog.askopenfilenames(**self.file_opt))
        self.current_files.sort(key=lambda x: x.lower())
        self.current_files = [pathlib.Path(path) for path in self.current_files]
        self.selected_files.delete(0, tkinter.END)
        if self.current_files:
            for current_file in self.current_files:
                self.selected_files.insert(tkinter.END, current_file.name)
            self.file_opt["initialdir"] = self.current_files[0].parent

    def files(self):
        return self.current_files or []


class SemTkExportSelector(tkinter.ttk.Frame):
    def __init__(self, root):
        tkinter.ttk.Frame.__init__(self, root)

        self.root = root
        self.label = tkinter.ttk.Label(self.root, text="select output format:")

        self.export_formats = ["default"] + sorted(
            sem.exporters.available_exporters()
        )
        self.export_combobox = tkinter.ttk.Combobox(self.root)
        self.export_combobox["values"] = self.export_formats
        self.export_combobox.current(0)

    def pack(self):
        self.label.pack()
        self.export_combobox.pack()

    def grid(self, row=0, column=0):
        x = row
        y = column
        self.label.grid(row=x, column=y)
        x += 1
        self.export_combobox.grid(row=x, column=y)
        x += 1
        return (x, y)

    def export_format(self):
        return self.export_combobox.get()


class SEMTkWapitiTrain(tkinter.ttk.Frame):
    def __init__(
        self,
        file_selector,
        workflow,
        annotation_name,
        lang,
        annotation_level=None,
        document_filter=None,
        top=None,
        main_frame=None,
        text="Algorithm-specific variables",
    ):
        if top:
            self.trainTop = top
        else:
            self.trainTop = tkinter.Toplevel()

        self.file_selector = file_selector
        self.workflow = workflow
        self.annotation_name = annotation_name
        self.lang = lang
        self.annotation_level = annotation_level or tkinter.StringVar(
            self.trainTop, value="top level"
        )
        self.document_filter = document_filter or tkinter.StringVar(
            self.trainTop, value="all documents"
        )

        self.current_train = sem.SEM_DATA_DIR / "train"
        if not self.current_train.exists():
            os.makedirs(self.current_train)

        self.main_frame = main_frame or self.trainTop
        self.trainTop.focus_set()
        self.CRF_algorithm_var = tkinter.StringVar(self.trainTop, value="rprop")
        self.CRF_l1_var = tkinter.StringVar(self.trainTop, value="0.5")
        self.CRF_l2_var = tkinter.StringVar(self.trainTop, value="0.0001")
        self.CRF_nprocs_var = tkinter.StringVar(self.trainTop, value="1")
        self.pattern_label_var = tkinter.StringVar(self.trainTop, value="")
        self.compact_var = tkinter.IntVar()
        self.compact_var.set(1)

        algsFrame = tkinter.ttk.LabelFrame(self.trainTop, text=text)
        algsFrame.pack(fill="both", expand="yes")

        crf_cur_row = 0

        tkinter.ttk.Button(algsFrame, text="pattern (optional)", command=self.select_file).grid(
            row=crf_cur_row, column=0, sticky=tkinter.W
        )
        self.pattern_label = tkinter.Label(algsFrame, textvariable=self.pattern_label_var)
        self.pattern_label.grid(row=crf_cur_row, column=1, sticky=tkinter.W)
        crf_cur_row += 1

        tkinter.Label(algsFrame, text="algorithm").grid(row=crf_cur_row, column=0, sticky=tkinter.W)
        CRF_algorithmValue = tkinter.ttk.Combobox(algsFrame, textvariable=self.CRF_algorithm_var)
        CRF_algorithmValue["values"] = ["rprop", "l-bfgs", "sgd-l1", "bcd", "rprop+", "rprop-"]
        CRF_algorithmValue.current(0)
        CRF_algorithmValue.grid(row=crf_cur_row, column=1)
        crf_cur_row += 1

        tkinter.Label(algsFrame, text="l1").grid(row=crf_cur_row, column=0, sticky=tkinter.W)
        CRF_algorithmValue = tkinter.Entry(algsFrame, textvariable=self.CRF_l1_var)
        CRF_algorithmValue.grid(row=crf_cur_row, column=1)
        crf_cur_row += 1

        tkinter.Label(algsFrame, text="l2").grid(row=crf_cur_row, column=0, sticky=tkinter.W)
        CRF_algorithmValue = tkinter.Entry(algsFrame, textvariable=self.CRF_l2_var)
        CRF_algorithmValue.grid(row=crf_cur_row, column=1)
        crf_cur_row += 1

        tkinter.Label(algsFrame, text="number of processors").grid(
            row=crf_cur_row, column=0, sticky=tkinter.W
        )
        CRF_nprocsValue = tkinter.ttk.Combobox(algsFrame, textvariable=self.CRF_nprocs_var)
        CRF_nprocsValue["values"] = list(range(1, multiprocessing.cpu_count() + 1))
        CRF_nprocsValue.current(0)
        CRF_nprocsValue.grid(row=crf_cur_row, column=1)
        crf_cur_row += 1

        tkinter.ttk.Checkbutton(
            algsFrame, text="compact model", variable=self.compact_var
        ).grid(row=crf_cur_row, column=0, sticky=tkinter.W)
        crf_cur_row += 1

        CRF_trainButton = tkinter.Button(algsFrame, text="train", command=self.trainCRF)
        CRF_trainButton.grid(row=crf_cur_row, column=0)
        crf_cur_row += 1

    def select_file(self, event=None):
        options = {}
        options["defaultextension"] = ".txt"
        options["filetypes"] = [("all files", ".*"), ("text files", ".txt")]
        options["initialdir"] = sem.SEM_DATA_DIR / "resources" / "patterns"
        options["parent"] = self.trainTop
        options["title"] = "Select pattern file."
        pattern = tkinter.filedialog.askopenfilename(**options)
        self.pattern_label_var.set(pattern)

    def algorithm(self):
        return self.CRF_algorithm_var.get()

    def l1(self):
        return float(self.CRF_l1_var.get())

    def l2(self):
        return float(self.CRF_l2_var.get())

    def nprocs(self):
        return int(self.CRF_nprocs_var.get())

    def pattern(self):
        return self.pattern_label_var.get()

    def compact(self):
        return bool(self.compact_var.get())

    def trainCRF(self, events=None):
        alg = self.algorithm()
        l1 = self.l1()
        l2 = self.l2()
        pattern = self.pattern()
        nprocs = self.nprocs()
        compact = self.compact()
        workflowfile = self.workflow.resource()
        export_format = "conll"
        pipeline, workflow_options, exporter, couples = sem.pipelines.load_workflow(
            workflowfile, force_format=export_format, pipeline_mode="train"
        )
        workflow_options = configparser.RawConfigParser()
        workflow_options.add_section("log")
        workflow_options.set("log", "log_level", "INFO")
        annotation_level = str2filter[self.annotation_level.get()]
        document_filter = str2docfilter[self.document_filter.get()]

        target_model = None
        pipes = [pipe for pipe in pipeline]
        trained_pipe = None
        for pipe in reversed(pipes):
            if isinstance(pipe, EnrichProcessor):
                pipe.mode = "train"
                self.annotation_name = pipe.informations.aentries[-1].name
                pipe.mode = "label"
                break
            elif isinstance(pipe, WapitiLabelProcessor):
                self.annotation_name = pipe.field
                target_model = pathlib.Path(pipe.model)
                trained_pipe = pipe
                break

        out_dir = None
        if target_model:
            out_dir = target_model.parent
            name = target_model.name
            try:
                os.makedirs(out_dir)
            except FileExistsError:
                pass

        timestamp = time.strftime("%Y%m%d%H%M%S")
        output_dir = self.current_train / timestamp
        if not output_dir.exists():
            os.makedirs(output_dir)

        train_file = output_dir / "train.conll"
        model_file = output_dir / "model.txt"
        # files can be selected from GUI or already provided (list)
        try:
            files = self.file_selector.files()
        except AttributeError:
            files = self.file_selector
        fields = []
        names = set()
        with open(train_file, "w", encoding="utf-8") as output_stream:
            for filename in files:
                document = sem.importers.load(
                    filename, encoding="utf-8", tagset_name=self.annotation_name
                )
                args = {
                    "workflow": None,
                    "infiles": [document],
                    "pipeline": pipeline,
                    "options": workflow_options,
                    "exporter": None,
                    "couples": None,
                }
                if isinstance(document, SEMCorpus):
                    for doc in document:
                        if doc.name in names:
                            sem.logger.warning("document %s already found, skipping", doc.name)
                            continue
                        elif not document_filter(doc, self.annotation_name):
                            sem.logger.warning("document %s has no annotations, skipping", doc.name)
                            continue
                        args["infiles"] = [doc]
                        doc = tagger(**args)[0]

                        if self.annotation_name is not None:
                            doc.add_to_corpus(self.annotation_name, filter=annotation_level)
                            doc.corpus.write(output_stream)
                        if not fields:
                            fields = doc.corpus.fields[:-1]
                else:
                    if document.name in names:
                        sem.logger.warning("document %s already found, skipping", document.name)
                        continue
                    elif not document_filter(document, self.annotation_name):
                        sem.logger.warning(
                            "document %s has no annotations, skipping", document.name
                        )
                        continue

                    document = tagger(**args)[0]

                    if self.annotation_name is not None:
                        document.add_to_corpus(self.annotation_name, filter=annotation_level)
                        document.corpus.write(output_stream)
                        if not fields:
                            fields = document.corpus.fields[:-1]

        pattern_file = output_dir / "pattern.txt"
        if pattern:
            shutil.copy(pattern, pattern_file)
        else:
            with open(pattern_file, "w", encoding="utf-8") as output_stream:
                output_stream.write("u\n\n")
                for i, field in enumerate(fields):
                    for shift in range(-2, 3):
                        output_stream.write("u:{0} {1:+d}=%x[{1},{2}]\n".format(field, shift, i))
                    output_stream.write("\n")
                output_stream.write("b\n")

        with open(output_dir / "config.txt", "w", encoding="utf-8") as output_stream:
            output_stream.write("algorithm\t{0}\n".format(alg))
            output_stream.write("l1\t{0}\n".format(l1))
            output_stream.write("l2\t{0}\n".format(l2))
            output_stream.write("number of processors\t{0}\n".format(nprocs))
            output_stream.write("compact\t{0}\n".format(compact))

        with open(pattern_file, "r") as input_stream:
            patterns = input_stream.read()
        model = WapitiModel(
            patterns=patterns, algo=alg, rho1=l1, rho2=l2, nthreads=nprocs, compact=compact
        )
        sequences = []
        for sequence in sem.importers.read_conll(train_file, "utf-8"):
            sequences.append(bytes(sequence.conll(range(len(sequence.keys()))) + "\n", "utf-8"))
        model.train(sequences)
        model.save(str(model_file))
        del model

        model_update_message = (
            "\n\nNo candidate location found, model update has to be done manually"
        )
        output_pipeline = sem.SEM_RESOURCE_DIR / "pipelines" / self.lang / workflowfile.stem
        try:
            output_pipeline.parent.mkdir(parents=True)
        except FileExistsError:
            pass
        if target_model:
            if target_model.exists():
                bname = pathlib.Path(name).stem
                ext = pathlib.Path(name).suffix
                backup_name = "{}.backup-{}{}".format(bname, timestamp, ext)
                dest = out_dir / backup_name
                sem.logger.info("creating backup file before moving: %s", dest)
                shutil.move(target_model, dest)
            sem.logger.info("trained model moved to: %s", str(out_dir))
            model_update_message = "\n\nTrained model moved to: {0}".format(out_dir)
            shutil.copy(model_file, target_model)
            pipeline.pipeline_mode = "label"
            trained_pipe.model = target_model
            sem.pipelines.save(pipeline, output_pipeline)
            pipeline.pipeline_mode = "train"
            pipeline_update_message = "\n\nPipeline saved to: {0}".format(output_pipeline)

        sem.logger.info("files are located in: %s", str(output_dir))
        tkinter.messagebox.showinfo(
            "training SEM",
            "Everything went ok!\n\nfiles are located in: {0}{1}{2}".format(
                output_dir, model_update_message, pipeline_update_message
            ),
        )

        if self.main_frame:
            self.main_frame.destroy()
        else:
            self.trainTop.destroy()


class SEMTkTrainInterface(tkinter.ttk.Frame):
    def __init__(self, documents, lang=None, workflow=None):
        self.documents = documents
        self._lang = lang
        self._workflow = workflow
        if not (workflow and lang):
            self._workflow = None
            self._lang = None
        self.workflow_selector = None
        self.lang_selector = None

        trainTop = tkinter.Toplevel()
        trainTop.focus_set()

        varsFrame = tkinter.ttk.LabelFrame(trainTop, text="Global variables")

        document_filter_label = tkinter.ttk.Label(varsFrame, text="document filter:")
        document_filter_var = tkinter.StringVar(varsFrame, value="all documents")
        document_filter = tkinter.ttk.Combobox(varsFrame, textvariable=document_filter_var)
        document_filter["values"] = sorted(str2docfilter.keys())
        document_filter.current(sorted(str2docfilter.keys()).index("all documents"))

        annotation_level_label = tkinter.ttk.Label(varsFrame, text="annotation level:")
        annotation_level_var = tkinter.StringVar(varsFrame, value="top level")
        annotation_level = tkinter.ttk.Combobox(varsFrame, textvariable=annotation_level_var)
        annotation_level["values"] = sorted(str2filter.keys())
        annotation_level.current(sorted(str2filter.keys()).index("top level"))

        directory = sem.SEM_DATA_DIR / "resources" / "workflow"
        if not self._workflow:
            self.workflow_selector = SemTkResourceSelector(
                varsFrame, directory, filter=lambda x: pathlib.Path(x).suffix == ".xml"
            )
        if not self._lang:
            self.lang_selector = SemTkLangSelector(varsFrame, directory)
            self.lang_selector.register(self.workflow_selector)

        algsFrame = tkinter.ttk.LabelFrame(trainTop, text="Algorithm-specific variables")

        notebook = tkinter.ttk.Notebook(algsFrame)
        frame1 = tkinter.ttk.Frame(notebook)
        frame2 = tkinter.ttk.Frame(notebook)
        notebook.add(frame1, text="CRF")
        notebook.add(frame2, text="NN")
        frame1.resource_dir = sem.SEM_DATA_DIR / "resources"

        varsFrame.pack(fill="both", expand="yes")
        vars_cur_row = 0
        document_filter_label.grid(row=vars_cur_row, column=0)
        vars_cur_row += 1
        document_filter.grid(row=vars_cur_row, column=0)
        vars_cur_row += 1
        annotation_level_label.grid(row=vars_cur_row, column=0)
        vars_cur_row += 1
        annotation_level.grid(row=vars_cur_row, column=0)
        vars_cur_row += 1
        if self.lang_selector:
            vars_cur_row, _ = self.lang_selector.grid(row=vars_cur_row, column=0)
        if self.workflow_selector:
            vars_cur_row, _ = self.workflow_selector.grid(row=vars_cur_row, column=0)

        for _ in range(5):
            tkinter.ttk.Separator(trainTop, orient=tkinter.HORIZONTAL).pack()

        algsFrame.pack(fill="both", expand="yes")

        notebook.pack()

        SEMTkWapitiTrain(
            self.documents,
            self.workflow,
            None,
            lang=self.lang,
            annotation_level=annotation_level_var,
            document_filter=document_filter_var,
            top=frame1,
            main_frame=trainTop,
            text="CRF-specific variables",
        )

    @property
    def workflow(self):
        return self._workflow or self.workflow_selector

    @property
    def lang(self):
        return self._lang or self.lang_selector.lang()


class SearchFrame(tkinter.ttk.Frame):
    def __init__(self, text, regexp=False, nocase=False):
        self.text = text
        self.pattern = tkinter.StringVar()
        self.regexp = tkinter.IntVar(int(regexp))
        self.nocase = tkinter.IntVar(int(nocase))
        self.prev_pattern = ""
        self.prev_regexp = regexp
        self.prev_nocase = nocase
        self.findings = []
        self.current = -1
        self.text.tag_config("search", background="yellow", foreground="black")
        bold_font = tkinter.font.Font(self.text)
        bold_font.configure(weight="bold")
        self.text.tag_config("search_current", font=bold_font)

    def clear_tags(self):
        self.text.tag_remove("search", "1.0", "end")
        self.text.tag_remove("search_current", "1.0", "end")

    def clear(self):
        self.clear_tags()
        self.prev_pattern = ""
        self.prev_regexp = self.regexp.get()
        self.prev_nocase = self.nocase.get()
        self.current = -1
        del self.findings[:]

    def find_in_text(self, event=None):
        find_in_text_top = tkinter.Toplevel()
        find_in_text_top.title("search")
        find_in_text_top.focus_set()
        matchesVar = tkinter.StringVar()

        def nxt(event=None):
            pattern = self.pattern.get()
            regexp = self.regexp.get()
            nocase = self.nocase.get()
            if (
                pattern != self.prev_pattern
                or regexp != self.prev_regexp
                or nocase != self.prev_nocase
            ):
                self.clear_tags()
                self.prev_pattern = pattern
                self.prev_regexp = regexp
                self.prev_nocase = nocase
                del self.findings[:]
                start = 1.0
                countVar = tkinter.StringVar()
                pos = self.text.search(
                    pattern,
                    start,
                    stopindex="end",
                    count=countVar,
                    regexp=self.regexp.get(),
                    nocase=self.nocase.get(),
                )
                while pos:
                    end = "{0} + {1}c".format(pos, countVar.get())
                    self.text.tag_add("search", pos, end)
                    self.findings.append((pos, end))
                    start = end
                    pos = self.text.search(
                        pattern,
                        start,
                        stopindex="end",
                        count=countVar,
                        regexp=self.regexp.get(),
                        nocase=self.nocase.get(),
                    )
                self.current = 1
            elif self.findings:
                prev = self.findings[self.current - 1]
                self.current = self.current + 1 if self.current < len(self.findings) else 1
                self.text.tag_remove("search_current", prev[0], prev[1])

            if self.findings:
                self.text.tag_add(
                    "search_current",
                    self.findings[self.current - 1][0],
                    self.findings[self.current - 1][1],
                )
                matchesVar.set("match {0} out of {1}".format(self.current, len(self.findings)))
                self.text.mark_set("insert", self.findings[self.current - 1][1])
                self.text.see("insert")
            else:
                matchesVar.set("no matches found")

        def cancel(event=None):
            self.clear()
            find_in_text_top.destroy()

        label1 = tkinter.Label(find_in_text_top, text="search for:")

        text = tkinter.ttk.Entry(find_in_text_top, textvariable=self.pattern)

        next_btn = tkinter.ttk.Button(find_in_text_top, text="next", command=nxt)
        cancel_btn = tkinter.ttk.Button(find_in_text_top, text="cancel", command=cancel)
        matches_found_lbl = tkinter.Label(find_in_text_top, textvariable=matchesVar)
        regexp_btn = tkinter.ttk.Checkbutton(
            find_in_text_top, text="regular expression", variable=self.regexp
        )
        nocase_btn = tkinter.ttk.Checkbutton(
            find_in_text_top, text="ignore case", variable=self.nocase
        )
        label1.grid(row=0, column=0)
        text.grid(row=0, column=1)
        regexp_btn.grid(row=1, column=0)
        nocase_btn.grid(row=1, column=1)
        next_btn.grid(row=2, column=0)
        cancel_btn.grid(row=2, column=1)
        matches_found_lbl.grid(row=3, column=0, columnspan=2)
        text.focus_set()

        find_in_text_top.bind("<Return>", nxt)
        find_in_text_top.bind("<Escape>", cancel)
        find_in_text_top.protocol("WM_DELETE_WINDOW", cancel)



# -*- coding: utf-8 -*-

"""
file: annotators.py

Description: some annotators that can be used in SEM.

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import pathlib
import functools

from sem.features import (directory_feature, xml2feat, multiword_dictionary)
from sem.CRF import Model as WapitiModel
from sem.util import check_model_available
from sem.storage import (compile_multiword, chunks_to_annotation, get_top_level)


class LexicaAnnotator:
    """An annotator that uses lexica to provide annotations."""

    __strategies = ("replace", "overriding", "non-overriding")

    def __init__(self, field, lexica, strategy="replace", *args, **kwargs):
        """The initialisation method of LexicaAnnotator.

        Parameters
        ----------
        lexica : function sem.storage.Sentence -> list[str]
            Lexica is a functools.partial of `sem.features.directory_feature` with multiple
            lexica that are either token or multiword features, used to provide annotations.
        field : str
            the name of the CoNLL field where annotations will be put. It is also the name
            of the annotation set where annotations will be stored.
        """
        self._field = field
        self._feature = lexica
        self._strategy = strategy
        if self._strategy not in LexicaAnnotator.__strategies:
            raise ValueError(f"Unsupported strategy: '{strategy}' {LexicaAnnotator.__strategies}")

    @classmethod
    def load(
        cls,
        field,
        features=None,
        xmllist=(),
        path=None,
        lexica=None,
        token_field="word",
        strategy="replace",
        **kwargs
    ):
        lexica = lexica or {}
        dirfeat = None  # directory_feature
        features = features or []
        if features:
            dirfeat = functools.partial(directory_feature, features=features)
            return LexicaAnnotator(field, dirfeat)
        elif xmllist:
            path = path or pathlib.Path(".")
            for child in xmllist:
                features.append(xml2feat(child, default_entry=token_field, path=path))
            dirfeat = functools.partial(directory_feature, features=features)
            return LexicaAnnotator(field, dirfeat)
        elif lexica:
            feats = []
            for tag_value, entries in lexica.items():
                appendice = f"-{tag_value}"
                trie = compile_multiword(entries)
                feat = functools.partial(
                    multiword_dictionary, trie=trie, y=token_field, appendice=appendice
                )
                feats.append(feat)
            dirfeat = functools.partial(directory_feature, features=feats)
            return LexicaAnnotator(field, dirfeat, strategy=strategy)
        else:
            raise ValueError(f"No data provided for loading {cls.__name__}.")

    def process_document(self, document, *args, **kwargs):
        tags = []
        for sequence in document.corpus:
            tagging = self._feature(sequence)
            try:
                existing = sequence.feature(self._field)
            except KeyError:
                existing = []
            tagging = self.merge(tagging, existing)
            sequence.add(tagging[:], self._field)
            tags.append(tagging[:])

        document.add_annotation_from_tags(tags, self._field, self._field)

    def merge(self, new, old):
        strategy = self._strategy
        if strategy == "replace":
            return new
        elif strategy == "overriding":
            source = chunks_to_annotation(new)
            candidates = chunks_to_annotation(old)
        elif strategy == "non-overriding":
            source = chunks_to_annotation(old)
            candidates = chunks_to_annotation(new)

        indices_new = [set(range(ann.lb, ann.ub)) for ann in source]
        for candidate in candidates:
            indices = set(range(candidate.lb, candidate.ub))
            if not any(indices & idxs for idxs in indices_new):
                source.add(candidate)
        source = get_top_level(source)
        tags = ["O" for _ in new]
        for tag in source:
            val = tag.value
            tags[tag.lb] = f"B-{val}"
            for i in range(tag.lb+1, tag.ub):
                tags[i] = f"I-{val}"
        return tags


class WapitiAnnotator:
    """An annotator that uses a (python implementation of) wapiti model to provide annotations."""

    def __init__(self, field, model):
        self._field = field
        self._model = model

    @classmethod
    def load(cls, field, location, input_encoding=None, *args, **kwargs):
        check_model_available(location)
        return (field, WapitiModel.from_wapiti_model(location, encoding=input_encoding))

    def process_document(
        self, document, annotation_name=None, annotation_fields=None, *args, **kwargs
    ):
        if annotation_name is None:
            annotation_name = str(self._field)

        tags = []
        for sequence in document.corpus:
            tagging, _, _ = self._model.tag_viterbi(sequence)
            tags.append(tagging[:])

        document.add_annotation_from_tags(tags, self._field, annotation_name)


def load(kind, *args, **kwargs):
    clss = __annotators[kind]
    return clss.load(*args, **kwargs)


__annotators = {
    "lexica": LexicaAnnotator,
    "wapiti": WapitiAnnotator
}


def get_annotator(name):
    return __annotators[name]



# -*- coding: utf-8 -*-

"""
file: annotate.py

Description: this module allows annotation of SEM documents.

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse
import time
from datetime import timedelta

import sem.annotators
import sem.logger
from sem.importers import read_conll
from sem.importers import conll_file
from sem.exporters import CoNLLExporter
from sem.processors import AnnotateProcessor


def main(argv=None):
    annotate(**vars(parser.parse_args(argv)))


def annotate(
    infile,
    outfile,
    annotator,
    token_field,
    field,
    ienc=None,
    oenc=None,
    enc="utf-8",
    log_level="WARNING",
    log_file=None,
):
    """
    Takes a CoNLL-formatted file and write another CoNLL-formatted file
    with additional features in it.

    Parameters
    ----------
    infile : str
        the CoNLL-formatted input file.
    outfile : str
        the CoNLL-formatted output file.
    mdl : str
        the wapiti model file.
    log_level : str or int
        the logging level.
    log_file : str
        if not None, the file to log to (does not remove command-line
        logging).
    """

    start = time.time()

    if log_file is not None:
        sem.logger.addHandler(sem.logger.file_handler(log_file))
    sem.logger.setLevel(log_level)

    ienc = ienc or enc
    oenc = oenc or enc
    annotator = AnnotateProcessor(
        infile=infile,
        outfile=outfile,
        annotator=annotator,
        token_field=token_field,
        field=field,
        ienc=ienc,
        oenc=oenc,
        enc=enc,
        log_level=log_level,
        log_file=log_file
    )

    length = -1
    fields = None
    for sentence in read_conll(infile, ienc):
        fields = fields or [str(i) for i in range(len(sentence.keys()))]
        if length == -1:
            length = len(fields)
        if length != len(sentence.keys()):
            raise ValueError(
                "{} has inconsistent number of columns, found {} and {}".format(
                    infile, length, len(sentence[0])
                )
            )

    document = conll_file(infile, fields, fields[0], encoding=ienc)

    annotator.process_document(document)

    exporter = CoNLLExporter()

    exporter.document_to_file(document, None, outfile, encoding=oenc)

    laps = time.time() - start
    sem.logger.info("done in %s", timedelta(seconds=laps))


parser = argparse.ArgumentParser("An annotation tool for SEM.")

parser.add_argument("infile", help="The input file (CoNLL format)")
parser.add_argument("outfile", help="The output file (CoNLL format)")
parser.add_argument("annotator", help="The name of the annotator")
parser.add_argument(
    "location",
    help="The location of the data used for annotator" " (model, folder with lexica, etc.).",
)
parser.add_argument("token_field", help="The token field (not always useful).")
parser.add_argument("field", help="The output field.")
parser.add_argument("--input-encoding", dest="ienc", help="Encoding of the input (default: UTF-8)")
parser.add_argument("--output-encoding", dest="oenc", help="Encoding of the input (default: UTF-8)")
parser.add_argument(
    "--encoding",
    dest="enc",
    default="UTF-8",
    help="Encoding of both the input and the output (default: UTF-8)",
)
parser.add_argument(
    "-l",
    "--log",
    dest="log_level",
    choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
    default="WARNING",
    help="Increase log level (default: critical)",
)
parser.add_argument("--log-file", dest="log_file", help="The name of the log file")



# -*- encoding: utf-8 -*-

"""
file: test_segmentation.py

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import unittest
import re

import sem.tokenisers


FR_TEXT = """18 mai j'ai en effet un p, porte-serviette qui est 1,2% plus cool que m. Trucmuche et mme. Machinchose.

Test: un mot qui finit par m, zoom. Et on regarde s'il est bien segmenté!

test2.

I
Vérification qu'une fin de ligne sans ponctuation est considérée comme une fin de phrase ou non.

Des     espaces multiples d'   orgines nombreuses.

Un test de ponctuation... Un autre test de ponctuation… Et une fin de ligne sans ponctuation

St-Jean peut causer des soucis de segmentation.

Vous pouvez m'écrire à yoa.dupont@gmail.com pour des questions au sujet de SEM, n'hésitez pas à visiter la page : http://www.lattice.cnrs.fr/sites/itellier/SEM.html!
"""  # noqa: E501
FR_TOKENS = [
    "18", "mai", "j'", "ai", "en", "effet", "un", "p", ",", "porte-serviette", "qui", "est", "1,2",
    "%", "plus", "cool", "que", "m.", "Trucmuche", "et", "mme.", "Machinchose", ".", "Test", ":",
    "un", "mot", "qui", "finit", "par", "m", ",", "zoom", ".", "Et", "on", "regarde", "s'", "il",
    "est", "bien", "segmenté", "!", "test", "2", ".", "I", "Vérification", "qu'", "une", "fin",
    "de", "ligne", "sans", "ponctuation", "est", "considérée", "comme", "une", "fin", "de",
    "phrase", "ou", "non", ".", "Des", "espaces", "multiples", "d'", "orgines", "nombreuses", ".",
    "Un", "test", "de", "ponctuation", "...", "Un", "autre", "test", "de", "ponctuation", "…",
    "Et", "une", "fin", "de", "ligne", "sans", "ponctuation", "St-Jean", "peut", "causer", "des",
    "soucis", "de", "segmentation", ".", "Vous", "pouvez", "m'", "écrire", "à",
    "yoa.dupont@gmail.com", "pour", "des", "questions", "au", "sujet", "de", "SEM", ",", "n'",
    "hésitez", "pas", "à", "visiter", "la", "page", ":",
    "http://www.lattice.cnrs.fr/sites/itellier/SEM.html", "!"
]
EN_TEXT = """I,  I've tried to improve Bob's coolness by 1.2%.
Test

test2.

I am
just checking if a newline is considered as an end of sentence or not.
"""


class TestSegmentation(unittest.TestCase):
    def test_get(self):
        default = sem.tokenisers.get_tokeniser("default")()  # noqa: F841
        fr = sem.tokenisers.get_tokeniser("fr")()  # noqa: F841
        en = sem.tokenisers.get_tokeniser("en")()  # noqa: F841

    def test_fr(self):
        content = FR_TEXT
        tokens_reference = FR_TOKENS[:]
        tokeniser = sem.tokenisers.FrenchTokeniser()
        token_spans = tokeniser.word_spans(content)
        tokens_guess = [content[s.lb: s.ub] for s in token_spans]

        self.assertEquals(tokens_guess, tokens_reference)

    def test_en(self):
        content = EN_TEXT
        tokeniser = sem.tokenisers.EnglishTokeniser()
        token_spans = tokeniser.word_spans(content)
        token_content = "".join([content[s.lb: s.ub] for s in token_spans])
        spaceless_content = re.sub(r"\s+", "", content)

        self.assertEquals(token_content, spaceless_content)  # no lost content


if __name__ == "__main__":
    unittest.main(verbosity=2)



"""
file: augment_wapiti_model.py

author: Yoann Dupont

MIT License

Copyright (c) 2021 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse
import json


__DEFAULT_PREFIXES = ("u:WordId_", "u:word ")


# This class allows to have a minimal CRF object that allows to easily augment the features.
# Unlike sem.CRF.Model, which was more made to load a Wapiti model that is not meant to be modified.
# Maybe later, this class will be replaced by sem.CRF.Model.
class CRF:
    def __init__(self):
        self.model_type = -1
        self.maxcol = -1
        self.autouni = 0
        self.patterns = []
        self.labels = []
        self.features = []
        self.theta = []
        self.offset = {}

    def weightsperfeat(self, feat):
        nlabels = len(self.labels)
        d = {b"u": nlabels, b"b": nlabels**2, b"*": nlabels+nlabels**2}
        return d[bytes([feat[0]])]

    def write(self, stream):
        nact = sum(f != 0.0 for f in self.theta)
        stream.write(b"#mdl#%i#%i\n" %(self.model_type, nact))

        stream.write(b"#rdr#%i/%i/%i\n" %(len(self.patterns), self.maxcol, self.autouni))
        for pattern in self.patterns:
            stream.write(b"%i:%s,\n" %(len(pattern), pattern))

        stream.write(b"#qrk#%i\n" %(len(self.labels)))
        for label in self.labels:
            stream.write(b"%i:%s,\n" %(len(label), label))

        stream.write(b"#qrk#%i\n" %(len(self.features)))
        for feature in self.features:
            stream.write(b"%i:%s,\n" %(len(feature), feature))

        for i, w in enumerate(self.theta):
            if w == 0.0:
                continue
            stream.write(b"%i=%s\n" %(i, bytes(float.hex(w), "utf-8")))

    @staticmethod
    def read(path):
        model = CRF()
        with open(path, "rb") as input_stream:
            header = next(input_stream).strip().split(b"#")
            model.model_type = int(header[2])
            nact = int(header[-1])

            header = next(input_stream).strip().split(b"#")
            npat, maxcol, autouni = header[-1].split(b"/")
            npat = int(npat)
            model.maxcol = int(maxcol)
            model.autouni = int(autouni)
            for i in range(npat):
                line = next(input_stream).strip().split(b":", 1)[1][:-1]
                model.patterns.append(line)

            header = next(input_stream).strip().split(b"#")
            nlabels = int(header[-1])
            for i in range(nlabels):
                line = next(input_stream).strip().split(b":", 1)[1][:-1]
                model.labels.append(line)

            header = next(input_stream).strip().split(b"#")
            nfeats = int(header[-1])
            numweights = 0
            weightsperfeat = {b"u": nlabels, b"b": nlabels**2, b"*": nlabels+nlabels**2}
            for i in range(nfeats):
                line = next(input_stream).strip().split(b":", 1)[1][:-1]
                model.features.append(line)
                model.offset[line] = numweights
                numweights += weightsperfeat[bytes([line[0]])]

            model.theta = [0.0 for _ in range(numweights)]
            for i in range(nact):
                line = next(input_stream).strip()
                idx, val = line.split(b"=", 1)
                idx = int(idx)
                model.theta[idx] = float.fromhex(val.decode("utf8"))
        return model


def main(argv=None):
    augment_wapiti_model(**vars(parser.parse_args(argv)))


def augment_wapiti_model(
    inputfile, configfile, outputfile, encoding="utf-8", substitution_prefixes=None
):
    mdl = CRF.read(inputfile)
    mdlfeats = set(mdl.features)
    prefixes = list(substitution_prefixes or __DEFAULT_PREFIXES)
    prefixes = [bytes(p, encoding) for p in prefixes]

    with open(configfile, encoding=encoding) as input_stream:
        config = json.load(input_stream)
    config["substitutions"] = [
        [bytes(p, encoding), bytes(r, encoding)] for p, r in config["substitutions"]
    ]
    config["additions"] = [
        [bytes(f, encoding), bytes(l, encoding), w] for f, l, w in config["additions"]
    ]

    add_feats = []
    for pattern, repl in config["substitutions"]:
        for feat in mdl.features:
            if not any(feat.startswith(prefix) for prefix in prefixes):
                continue
            if pattern not in feat:
                continue
            off1 = mdl.offset[feat]
            addition = feat.replace(pattern, repl)
            if addition in mdlfeats:
                continue
            add_feats.append(addition)
            mdl.offset[addition] = len(mdl.theta)
            mdl.theta.extend(mdl.theta[off1: off1+mdl.weightsperfeat(feat)])
    for feat, label, weight in config["additions"]:
        if feat in mdlfeats:
            continue
        weights = [0.0 for _ in range(mdl.weightsperfeat(feat))]
        weights[mdl.labels.index(label)] = weight
        add_feats.append(feat)
        mdl.offset[feat] = len(mdl.theta)
        mdl.theta.extend(weights)
    mdl.features.extend(add_feats)

    with open(outputfile, "wb") as output_stream:
        mdl.write(output_stream)


parser = argparse.ArgumentParser(
    'Add more features in a model by substitution or adding entirely new features.'
    ' This is specially useful when one wants to add features for words that were not seen or for'
    ' handling special characters.\n'
    '\n'
    'The configuration file is a json file with two keys: substitutions and additions.'
    ' Here is an example of such a json:\n'
    '\n'
    '{\n'
    '    "substitutions": [\n'
    '        ["\'", "ʼ"],\n'
    '        ["\'", "’"]\n'
    '    ],\n'
    '    "additions": [\n'
    '        ["u:WordId_+0=«", "PONCT", 32.0],\n'
    '        ["u:WordId_+0=»", "PONCT", 32.0],\n'
    '        ["u:WordId_+0=–", "PONCT", 32.0],\n'
    '        ["u:WordId_+0=—", "PONCT", 32.0],\n'
    '        ["u:WordId_+0=•", "PONCT", 32.0]\n'
    '    ]\n'
    '}\n'
)
parser.add_argument("inputfile", help="The input wapiti model to augment.")
parser.add_argument("configfile", help="The json file where substitutions and additions are given.")
parser.add_argument("outputfile", help="The output (augmented) wapiti model.")
parser.add_argument(
    "-e", "--encoding", default="utf-8", help="The encoding of configfile. (default: %(default)s)"
)
parser.add_argument(
    "--substitution-prefixes",
    nargs="*",
    default=__DEFAULT_PREFIXES,
    help="The allowed feature prefixes for substitutions. (default: %(default)s)"
)



# -*- coding: utf-8 -*-

"""
file: util.py

Description: a file for utilitarian functionalities.

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import pathlib
import re
import tarfile

from contextlib import contextmanager
from io import open, StringIO

import sem.logger


def find_suggestions(target, candidates, case_sensitive=True):
    trgt = target if case_sensitive else target.lower()
    suggestions = []
    for candidate in candidates if case_sensitive else [cand.lower() for cand in candidates]:
        shortest = candidate if len(candidate) < len(trgt) else trgt
        longest = trgt if shortest == candidate else candidate
        if shortest in longest:
            suggestions.append(candidate)
    return suggestions


def ranges_to_set(ranges, length, include_zero=False):
    """
    Returns a set of integers based of ranges over a reference list. Ranges are
    inclusive both in the lower and upper bound, unlike in python where they are
    inclusive only in the lower but not the upper bound.

    Parameters
    ----------
    ranges : str
        a list of ranges represented by a string. Each range is comma separated.
        A range is a couple of integers colon separated.
        If the index 0 in not present in the list, it is automatically added.
    length : int
        the reference length to use in case of negative indexing.
    """
    result = set()

    for current_range in ranges.split(","):
        if ":" in current_range:
            lo, hi = [int(i) for i in current_range.split(":")]
        else:
            lo = int(current_range)
            hi = lo

        if hi < lo:
            continue

        if lo < 0:
            lo = length + lo
        if hi < 0:
            hi = length + hi

        for i in range(lo, hi + 1):
            result.add(i)

    if include_zero:
        result.add(0)

    return result


def correct_pos_tags(tags):
    """
    Correct POS tags following the "tag _tag" scheme to remove impossible transitions.
    """
    corrected = [[level2 for level2 in level1] for level1 in tags]
    corrections = 0
    for i in range(len(corrected)):
        value = ""
        j = len(corrected[i]) - 1
        while j >= 0:
            if value != "":
                if corrected[i][j][0] == "_":
                    if corrected[i][j][1:] != value:
                        corrected[i][j] = "_{}".format(value)
                        corrections += 1
                else:
                    if corrected[i][j][0] != "_" and corrected[i][j] != value:
                        corrected[i][j] = value
                        corrections += 1
                    value = ""
            elif corrected[i][j][0] == "_":
                value = corrected[i][j][1:]
            j -= 1
    return corrected


def check_model_available(model):
    path = pathlib.Path(model)
    if not path.exists():
        targz = pathlib.Path(str(path) + ".tar.gz")
        if targz.exists():
            sem.logger.info("Model not extracted, extracting {0}".format(str(targz.resolve())))
            with tarfile.open(targz, "r:gz") as tar:
                tar.extractall(targz.parent)
        else:
            raise FileNotFoundError("Cannot find model file: {0}".format(model))


def strip_html(html, keep_offsets=False):
    """
    Take an str containing HTML data and strip the HTML markup.

    Parameters
    ----------
    html : str
        the HTML data (str) to strip.
    keep_offsets : bool [False]
        if True, markups are replaced by spaces, otherwise replace by single space.
    """

    def replace_same_size(m):
        return " " * (m.end() - m.start())

    preprocessed_html = html[:]
    to_remove = []
    to_remove.append(re.compile("<script.*?>.+?</script>", re.M + re.U + re.DOTALL))
    to_remove.append(re.compile("<head>.+?</head>", re.M + re.U + re.DOTALL))
    to_remove.append(re.compile("<nav.*?>.+?</nav>", re.M + re.U + re.DOTALL))
    for remove in to_remove:
        preprocessed_html = remove.sub(replace_same_size, preprocessed_html)

    to_keep = []
    to_keep.append(re.compile("<h[1][^>]*?>.+?</h[0-9]>", re.M + re.U + re.DOTALL))
    to_keep.append(re.compile("<p.*?>.+?</p>", re.M + re.U + re.DOTALL))

    parts = []
    s_e = []
    for keep in to_keep:
        for finding in keep.finditer(html):
            s_e.append([finding.start(), finding.end()])
    s_e.sort(key=lambda x: (x[0], -x[1]))
    ref = s_e[0]
    i = 1
    while i < len(s_e):
        if ref[0] <= s_e[i][0] and s_e[i][1] <= ref[1]:
            del s_e[i]
            i -= 1
        elif ref[1] <= s_e[i][0]:
            ref = s_e[i]
        i += 1

    if keep_offsets:
        non_space = re.compile("[^ \n\r]")
        parts.append(" " * s_e[0][0])
    else:
        non_space = re.compile("[^ \n\r]+")
    for i in range(len(s_e)):
        if i > 0:
            parts.append(non_space.sub(" ", preprocessed_html[s_e[i - 1][1]: s_e[i][0]]))
        parts.append(preprocessed_html[s_e[i][0]: s_e[i][1]])
    stripped_html = "".join(parts)

    tag = re.compile("<.+?>", re.U + re.M + re.DOTALL)
    repl = replace_same_size if keep_offsets else ""
    if keep_offsets:
        stripped_html = (
            tag.sub(repl, stripped_html).replace("&nbsp;", "      ").replace("&#160;", "      ")
        )
    else:
        stripped_html = tag.sub(repl, stripped_html).replace("&nbsp;", " ").replace("&#160;", " ")
        stripped_html = re.sub("\n[ \t]+", "\n", stripped_html)
        stripped_html = re.sub("[ \t]+", " ", stripped_html)
        stripped_html = re.sub("\n{2,}", "\n\n", stripped_html)

    return stripped_html


def str2bool(s):
    """
    Return a boolean value from string.

    Parameters
    ----------
    s : str
        the string value to convert to boolean.
    """

    s = s.lower()
    res = {"yes": True, "y": True, "true": True, "no": False, "n": False, "false": False}.get(
        s, None
    )
    if res is None:
        raise ValueError('Cannot convert to boolean: "{0}"'.format(s))
    return res


def longest_common_substring(a, b, casesensitive=True, lastchance=False):
    """
    A backtrack algorithm to find the best suited match between a long and
    a short form.
    It will generate a list of candidate solutions that have exactly the
    same number of matching tokens as the LCS algorithm found.

    """
    if len(a) < len(b):
        # left should be the longest sequence, but here right is.
        # We invert them so we do not have to deal with the mirroring of the checks.
        # However, in the end, the indexes will be switched between the two tokens,
        # so we put them back in place.
        return [
            [(y, x) for (x, y) in solution]
            for solution in longest_common_substring(b, a, casesensitive)
        ]
    if not casesensitive:
        # simple and dirty python to easily simulate case insensitivity.
        # Even though it works for most cases, it will not for some specific
        # unicode characters which I do not recall.
        return longest_common_substring(a.lower(), b.lower(), casesensitive=True)

    lengths = [[0 for j in range(len(b) + 1)] for i in range(len(a) + 1)]
    for i, x in enumerate(a):
        for j, y in enumerate(b):
            if x == y:
                lengths[i + 1][j + 1] = lengths[i][j] + 1
            else:
                lengths[i + 1][j + 1] = max(lengths[i + 1][j], lengths[i][j + 1])

    solutions = set()

    def bcktrck(x, y, current):
        """
        The recursive algorithm that finds solutions by backtracking
        """
        if x == 0 or y == 0:
            if current == []:  # empty solution
                return
            if len(current) != lengths[-1][-1]:
                return

            z, t = current[0]
            # we start in the middle of a word of the long form
            if z > 0 and a[z - 1].isalpha():
                return
            # we do not take the first character of the short form
            if t != 0:
                return

            solutions.add(tuple(current))
            return
        if a[x - 1] == b[y - 1]:
            try:
                a_x2_ok = a[x - 2].isalpha()
            except Exception:
                a_x2_ok = False

            if current != []:
                if a[current[0][0] - 1].isalpha():
                    bcktrck(x - 1, y, current)
                    return
                # generated more than one token, and splitted one of them
                if " " in a[x - 1: current[0][0]].strip() and a_x2_ok:
                    bcktrck(x - 1, y, current)
                    return
            else:
                # we cut somewhere after a hyphen
                if "-" in a[x - 1:] and not a[x - 1] == "-":
                    if lengths[x - 1][y] == lengths[x][y]:
                        bcktrck(x - 1, y, current)
                    return
                # generated more than one token *and* splitted one of them
                if " " in a[x - 1:].strip() and (x > 1 and a_x2_ok):
                    if lengths[x - 1][y] == lengths[x][y]:
                        bcktrck(x - 1, y, current)
                    return

            bcktrck(x - 1, y - 1, [(x - 1, y - 1)] + current)
            if lengths[x][y] == lengths[x - 1][y]:
                bcktrck(x - 1, y, current)
            if lengths[x][y] == lengths[x][y - 1]:
                bcktrck(x, y - 1, current)
        elif lengths[x][y] == lengths[x - 1][y]:
            bcktrck(x - 1, y, current)
        elif lengths[x][y] == lengths[x][y - 1]:
            bcktrck(x, y - 1, current)

    try:
        bcktrck(len(a), len(b), [])
        return list(solutions)
    except RuntimeError:  # maximum recursion depth
        return []


@contextmanager
def r_open(source, encoding="utf-8"):
    """Open in read mode either a file from the filename or a string."""

    src = None
    try:
        src = open(source, "r", encoding=encoding, newline="")
    except Exception:
        try:
            s = source.decode(encoding)
        except UnicodeDecodeError:
            s = source
        except UnicodeEncodeError:
            s = source
        except AttributeError:
            s = source
        src = StringIO(s)
    finally:
        yield src
        try:
            src.close()
        except AttributeError:
            pass


def read_chunks(source, size=1024, encoding="utf-8"):
    """Read a source by chunks of a given size."""

    with r_open(source, encoding) as input_stream:
        buff = input_stream.read(size)
        while buff:
            yield buff
            buff = input_stream.read(size)



# -*- coding:utf-8 -*-

"""
file: segmentation.py

Description: performs text segmentation according to given tokeniser.

author: Yoann Dupont

MIT License

Copyright (c) 2018 Yoann Dupont

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import argparse
import pathlib

from sem.storage import Document
import sem.logger
from sem.processors import SegmentationProcessor


def main(argv=None):
    segmentation(**vars(parser.parse_args(argv)))


def segmentation(
    infile,
    tokeniser_name,
    outfile,
    output_format,
    ienc=None,
    oenc=None,
    enc="utf-8",
    log_level="WARNING",
    log_file=None,
):
    if log_file is not None:
        sem.logger.addHandler(sem.logger.file_handler(log_file))
    sem.logger.setLevel(log_level)

    ienc = ienc or enc
    oenc = oenc or enc
    segmenter = SegmentationProcessor(tokeniser_name)
    with open(infile, "r", encoding=ienc) as input_stream:
        document = Document(
            pathlib.Path(infile).name,
            content=input_stream.read().replace("\r", ""),
        )
    segmenter.process_document(document)
    tokens_spans = document.segmentation("tokens")
    sentence_spans = document.segmentation("sentences")
    joiner = "\n" if output_format == "vector" else " "
    content = document.content
    with open(outfile, "w", encoding=oenc) as output_stream:
        for sentence in sentence_spans:
            sentence_token_spans = tokens_spans[sentence.lb: sentence.ub]
            sentence_tokens = [content[s.lb: s.ub] for s in sentence_token_spans]
            output_stream.write(joiner.join(sentence_tokens))
            if output_format == "vector":
                output_stream.write("\n")
            output_stream.write("\n")


parser = argparse.ArgumentParser(
    "Segments the textual content of a sentence into tokens."
    " They can either be outputted line per line or in a vectorised format."
)

parser.add_argument("infile", help="The input file (raw text)")
parser.add_argument("tokeniser_name", help="The name of the tokeniser to import")
parser.add_argument("outfile", help="The output file")
parser.add_argument(
    "--output-format",
    choices=("line", "vector"),
    default="vector",
    help="The output format (default: %(default)s)",
)
parser.add_argument("--input-encoding", dest="ienc", help="Encoding of the input (default: UTF-8)")
parser.add_argument("--output-encoding", dest="oenc", help="Encoding of the input (default: UTF-8)")
parser.add_argument(
    "--encoding",
    dest="enc",
    default="UTF-8",
    help="Encoding of both the input and the output (default: UTF-8)",
)
parser.add_argument(
    "-l",
    "--log",
    dest="log_level",
    choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
    default="WARNING",
    help="Increase log level (default: %(default)s)",
)
parser.add_argument("--log-file", dest="log_file", help="The name of the log file")
